#----------------------------------------------------------------------------------------------
#  Copyright (c) Microsoft Corporation. All rights reserved.
#  Licensed under the MIT License. See License.txt in the project root for license information.
#----------------------------------------------------------------------------------------------

from __future__ import absolute_import

import os
import gzip
from six import string_types as _string_types
import paddle.v2 as paddle
import paddle.trainer_config_helpers.layers as layers
import numpy as np
from mmdnn.conversion.paddle.paddle_graph import PaddleGraph
import mmdnn.conversion.common.IR.graph_pb2 as graph_pb2
from mmdnn.conversion.common.IR.graph_pb2 import NodeDef, GraphDef, DataType
from mmdnn.conversion.common.DataStructure.parser import Parser
from mmdnn.conversion.common.utils import *


class PaddleParser(Parser):

    dtype_map = {
        "float16" : graph_pb2.DT_FLOAT16,
        "float32" : graph_pb2.DT_FLOAT32,
        "float64" : graph_pb2.DT_FLOAT64,
        "int16"   : graph_pb2.DT_INT16,
        "int32"   : graph_pb2.DT_INT32,
        "int64"   : graph_pb2.DT_INT64,
        "uint8"   : graph_pb2.DT_UINT8,
        "uint16"  : graph_pb2.DT_UINT16
    }

    activation_map = {
        "relu"          : "Relu",
        'softmax'       : "Softmax",
        'sigmoid'       : "Sigmoid",
        "tanh"          : "Tanh",
        "elu"           : "Elu",
        "relu6"         : "Relu6",
        'softplus'      : 'Softplus',
        'softsign'      : 'Softsign',
        'hard_sigmoid'  : 'HardSigmoid'
    }

    layer_map = {
        "data"          : "InputLayer",
        "exconv"        : "Conv",
        "addto"         : "Add",
        "batch_norm"    : "BatchNormalization",
        "pool"          : "Pooling",
        "fc"            : "Dense",
        "norm"          : "LRN",


    }


    def _load_model(self, model_network_path, model_weight_path):
        """Load a paddle model from disk

        Parameters
        ----------
        model_network_path: str
            Path where the model network path is (json file)

        model_weight_path: str
            Path where the model network weights are (hd5 file)

        Returns
        -------
        model: A paddle model
        """
        from paddle.proto import ModelConfig_pb2
        from mmdnn.conversion.common.IR.IR_graph import load_protobuf_from_file

        loaded_model = ModelConfig_pb2.ModelConfig()
        load_protobuf_from_file(loaded_model, model_network_path)

        if model_weight_path:
            if os.path.isfile(model_weight_path):
                parameters = paddle.parameters.Parameters.from_tar(gzip.open(model_weight_path, 'r'))
                self.weight_loaded = True
                print("Network file [{}] and [{}] is loaded successfully.".format(model_network_path, model_weight_path))

            else:
                print("Warning: Weights File [%s] is not found." % (model_weight_path))

        return loaded_model, parameters

    @property
    def src_graph(self):
        return self.paddle_graph


    def __init__(self, model):
        super(PaddleParser, self).__init__()

        if isinstance(model, tuple):
            model_network_path, model_weight_path = model

        # Build network graph
        model, parameters = self._load_model(model_network_path, model_weight_path)
        self.paddle_graph = PaddleGraph(model)
        self.paddle_graph.build()
        self.parameters = parameters
        self.shape_dict = dict()





    def gen_IR(self):

        for layer in self.paddle_graph.topological_sort:
            current_node = self.paddle_graph.get_node(layer)
            node_type = PaddleParser.layer_map[current_node.type]
            if hasattr(self, "rename_" + node_type):
                func = getattr(self, "rename_" + node_type)
                func(current_node)
            else:
                print("PaddleParser has not supported operator [%s]." % (node_type))
                self.rename_UNKNOWN(current_node)



    @staticmethod
    def _set_output_shape(source_node, IR_node, output_shapes):
        shape = graph_pb2.TensorShape()
        for output_shape in output_shapes:
            new_dim = shape.dim.add()
            new_dim.size = output_shape
        IR_node.attr["_output_shapes"].list.shape.extend([shape])


    @staticmethod
    def _copy_and_reop(source_node, IR_node, new_op = None):
        IR_node.name = source_node.name.lstrip('_')
        IR_node.op = source_node.type if new_op == None else new_op

        if hasattr(source_node.layer, "dtype"):
            IR_node.attr["dtype"].type = PaddleParser.dtype_map[source_node.layer.dtype]

        # PaddleParser._set_output_shape(source_node, IR_node)


    @staticmethod
    def _copy_shape(source_node, target_node, output_shapes):
        for dim in output_shapes:
            new_dim = target_node.attr["shape"].shape.dim.add()
            new_dim.size =  dim


    @staticmethod
    def _convert_dataformat(source_node, target_node):
        if source_node.keras_layer.data_format == 'channels_last':
            target_node.attr["data_format"].s = "NHWC"
        elif source_node.keras_layer.data_format == 'channels_first':
            target_node.attr["data_format"].s = "NCHW"
        else:
            print("Warning: [%s] don't have data format info." % (source_node.keras_layer.name))




    def _defuse_activation(self, source_node):
        src_spec = source_node.layer

        IR_node = self.IR_graph.node.add()
        IR_node.name = source_node.real_name.lstrip('_') + "_activation"
        IR_node.op = PaddleParser.activation_map[src_spec.active_type.encode()]
        IR_node.input.append(source_node.real_name.lstrip('_'))

        source_node.real_name = IR_node.name
        return IR_node



    def _convert_merge(self, source_node, new_name = None):
        IR_node = self.IR_graph.node.add()

        # name, op
        PaddleParser._copy_and_reop(source_node, IR_node, new_name)

        # input edge
        self.convert_inedge(source_node, IR_node)

        # For concat axis
        if hasattr(source_node.layer, 'axis'):
            IR_node.attr['axis'].i = -1
        return IR_node



    def rename_UNKNOWN(self, source_node):
        print (source_node.layer.get_config())

        # only for training
        IR_node = self.IR_graph.node.add()

        # name, op
        PaddleParser._copy_and_reop(source_node, IR_node)

        # input edge
        self.convert_inedge(source_node, IR_node)


    def rename_Conv(self, source_node):
        IR_node = self.IR_graph.node.add()

        # input edge
        self.convert_inedge(source_node, IR_node)

        # layer and spec
        conv_spec = source_node.layer

        spec = conv_spec.inputs[0].conv_conf

        # width <=> x or height <=> y
        width = spec.filter_size
        height = spec.filter_size_y if spec.HasField('filter_size_y') else spec.filter_size
        inputchannel = spec.channels
        outputchannel = conv_spec.num_filters
        stride_x = spec.stride
        stride_y = spec.stride_y if spec.HasField('stride_y') else stride_x
        padding_x = spec.padding
        padding_y = spec.padding_y if spec.HasField('padding_y') else padding_x
        dilation_x = spec.dilation
        dilation_y = spec.dilation_y if spec.HasField('dilation_y') else dilation_x
        output_x = spec.output_x
        output_y = spec.output_y if spec.HasField('output_y') else output_x
        input_x = spec.img_size
        input_y = spec.img_size_y if spec.HasField('img_size_y') else input_x


        # output shape
        output_shapes = [-1, output_y, output_x, outputchannel]
        self.shape_dict[source_node.name] = output_shapes
        PaddleParser._set_output_shape(source_node, IR_node, output_shapes)


        kwargs = dict()

        if conv_spec.type == 'exconv' or 'cudnn_conv':
            # name, op
            PaddleParser._copy_and_reop(source_node, IR_node, "Conv")
        else:
            kwargs['isDeconvolution'] = True
            PaddleParser._copy_and_reop(source_node, IR_node, "ConvTranspose")


        w_name = conv_spec.inputs[0].input_parameter_name
        w = self.parameters.get(w_name)


        self.set_weight(IR_node.name, 'weights', w.reshape([outputchannel, inputchannel, height, width]).transpose([ 2, 3, 1, 0]))

        #  it should be in the shape of height x width x inputchannel x outputchannel

        # use_bias: TODO
        kwargs['use_bias'] = False
        if conv_spec.HasField('bias_parameter_name'):
            bias_name = conv_spec.bias_parameter_name
            bias = self.parameters.get(bias_name).squeeze()
            self.set_weight(IR_node.name, "bias", bias)
            kwargs['use_bias'] = True



        kwargs['kernel_shape'] = [height, width, inputchannel, outputchannel]



        # pad_dim
        pad_dim = [0, 0, padding_x, padding_y, padding_x, padding_y, 0, 0]

        # fail report because of auto_pad
        # if dilation_x == 1 and dilation_y == 1:
        #     if output_x * stride_x == input_x and output_y * stride_y == input_y:
        #         auto_pad = "SAME"
        #         kwargs['auto_pad'] = auto_pad
        #     elif output_x * stride_x == input_x - width + 1 and output_y * stride_y == input_y - height + 1:
        #         auto_pad = "VALID"
        #         kwargs['auto_pad'] = auto_pad

        if input_x == output_x and input_y == output_y:
            auto_pad = "SAME"
        else:
            auto_pad = "SAME"

        pad_dim = convert_tf_pad_to_onnx(pad_dim)
        kwargs['pads'] = pad_dim

        kwargs['group'] = spec.groups

        kwargs['dilation'] = [1, dilation_x, dilation_y, 1]

        kwargs['strides'] = [1, stride_x, stride_y, 1]

        assign_IRnode_values(IR_node, kwargs)

        # defuse the activation layer

        if conv_spec.HasField('active_type') and  conv_spec.active_type != '':
            IR_node_act = self._defuse_activation(source_node)
            PaddleParser._set_output_shape(source_node, IR_node_act, output_shapes)


    def rename_BatchNormalization(self, source_node):
        IR_node = self.IR_graph.node.add()

        # name, op
        PaddleParser._copy_and_reop(source_node, IR_node, "BatchNorm")

        # input edge
        self.convert_inedge(source_node, IR_node)

        # layer and spec
        bn_spec = source_node.layer



        # output shape
        if  bn_spec.inputs[0].HasField("image_conf"):
            img_conf = bn_spec.inputs[0].image_conf
            output_x = img_conf.img_size
            output_y = img_conf.img_size_y if img_conf.HasField('img_size_y') else output_x
            outputchannel = img_conf.channels

            output_shapes = [-1, output_y, output_x, outputchannel]
            self.shape_dict[source_node.name] = output_shapes
            PaddleParser._set_output_shape(source_node, IR_node, output_shapes)


        IR_node.attr['scale'].b = True
        IR_node.attr['bias'].b = bn_spec.HasField('bias_parameter_name')

        w_name = bn_spec.inputs[0].input_parameter_name
        mean_name = bn_spec.inputs[1].input_parameter_name
        var_name = bn_spec.inputs[2].input_parameter_name
        bias_name = bn_spec.bias_parameter_name

        gamma = self.parameters.get(w_name)
        mean = self.parameters.get(mean_name)
        variance = self.parameters.get(var_name)
        beta = self.parameters.get(bias_name)

        # channels_first, then axis = 1
        IR_node.attr['axis'].i = -1

        # epsilon
        IR_node.attr['epsilon'].f = bn_spec.epsilon

        # compute adjusted parameters
        # Reference: parameter transformation https://github.com/apple/coremltools/issues/153
        f = 1.0 / np.sqrt(variance +  bn_spec.epsilon)
        gamma1 = gamma*f
        beta1 = beta - gamma*mean*f
        mean[:] = 0.0 #mean
        variance[:] = 1.0 - .00001 #stddev

        # convert type because of tensorflow
        gamma1 = gamma1.astype(np.float32)
        beta1 = beta1.astype(np.float32)
        mean = mean.astype(np.float32)
        variance = variance.astype(np.float32)

        # flatten
        gamma1 = gamma1.flatten()
        beta1 = beta1.flatten()
        mean = mean.flatten()
        variance = variance.flatten()



        if IR_node.attr['scale'].b:
            self.set_weight(IR_node.name, "scale", gamma1)

        if IR_node.attr['bias'].b:
            self.set_weight(IR_node.name, "bias", beta1)

        # mean
        self.set_weight(IR_node.name, "mean", mean)

        # var
        self.set_weight(IR_node.name, "var", variance)

        # defuse the activation layer

        if bn_spec.HasField('active_type') and  bn_spec.active_type != '':
            IR_node_act = self._defuse_activation(source_node)
            if  bn_spec.inputs[0].HasField("image_conf"):
                PaddleParser._set_output_shape(source_node, IR_node_act, output_shapes)



    def rename_Pooling(self, source_node):
        IR_node = self.IR_graph.node.add()

        # name, op
        PaddleParser._copy_and_reop(source_node, IR_node, "Pool")

        # input edge
        self.convert_inedge(source_node, IR_node)

        # layer and spec
        pool_spec = source_node.layer
        spec = pool_spec.inputs[0].pool_conf



        # assert False
        kwargs = dict()

        if spec.pool_type == 'max-projection':
            kwargs['pooling_type'] = 'MAX'
        elif spec.pool_type == 'avg-projection':
            kwargs['pooling_type'] = 'AVG'
        else:
            kwargs['pooling_type'] = 'MAX'



        width = spec.size_x
        height = spec.size_y if spec.HasField('size_y') else width
        channel = spec.channels
        stride_x = spec.stride
        stride_y = spec.stride_y if spec.HasField('stride_y') else stride_x
        padding_x = spec.padding
        padding_y = spec.padding_y if spec.HasField('padding_y') else padding_x
        output_x = spec.output_x
        output_y = spec.output_y if spec.HasField('output_y') else output_x
        input_x = spec.img_size
        input_y = spec.img_size_y if spec.HasField('img_size_y') else input_x


        # output shape
        output_shapes = [-1, output_y, output_x, channel]
        self.shape_dict[source_node.name] = output_shapes
        PaddleParser._set_output_shape(source_node, IR_node, output_shapes)


        kwargs['global_pooling'] = False

        kwargs['strides'] = [1, stride_x, stride_y, 1]
        kwargs['kernel_shape'] = [1, width, height, 1]

        # pad_dim
        pad_dim = [0, 0, padding_x, padding_y, padding_x, padding_y, 0, 0]


        # padding mode
        # If padding == "SAME": output_spatial_shape[i] = ceil(input_spatial_shape[i] / strides[i])
        # If padding == "VALID": output_spatial_shape[i] = ceil((input_spatial_shape[i] - (spatial_filter_shape[i]-1) * dilation_rate[i]) / strides[i]).

        if output_x * stride_x == input_x and output_y * stride_y == input_y:
            auto_pad = "SAME"
            kwargs['auto_pad'] = auto_pad
        elif output_x * stride_x == input_x - width + 1 and output_y * stride_y == input_y - height + 1:
            auto_pad = "VALID"
            kwargs['auto_pad'] = auto_pad

        pad_dim = convert_tf_pad_to_onnx(pad_dim)
        kwargs['pads'] = pad_dim



        assign_IRnode_values(IR_node, kwargs)

        if pool_spec.HasField('active_type') and  pool_spec.active_type != '':
            IR_node_act = self._defuse_activation(source_node)
            PaddleParser._set_output_shape(source_node, IR_node_act, output_shapes)


    def rename_Dense(self, source_node):
        IR_node = self.IR_graph.node.add()

        # name, op
        PaddleParser._copy_and_reop(source_node, IR_node, "FullyConnected")

        # input edge
        self.convert_inedge(source_node, IR_node)

        # layer and spec
        fc_spec = source_node.layer


        # units
        IR_node.attr['units'].i = fc_spec.size


        # output shape
        output_shapes = [-1, fc_spec.size]
        self.shape_dict[source_node.name] = output_shapes
        PaddleParser._set_output_shape(source_node, IR_node, output_shapes)


        # use_bias
        IR_node.attr['use_bias'].b = fc_spec.HasField('bias_parameter_name')

        w_name = fc_spec.inputs[0].input_parameter_name
        bias_name = fc_spec.bias_parameter_name

        w = self.parameters.get(w_name)

        bias = self.parameters.get(bias_name).flatten()

        # Kit weight tranpose
        # weight: N x M -> C x H x W x M -> H x W x C x M -> N x M
        if self.weight_loaded:
            parent = self.src_graph.get_parent(source_node.name, [0])
            if len(self.shape_dict[parent.name]) == 4:
                #
                original_shape = w.shape
                channel_first_list = self.shape_dict[parent.name][1:]
                dim = len(channel_first_list) + 1
                weight = w.reshape(channel_first_list + [original_shape[1]])
                assert dim > 2
                weight = weight.transpose(list(range(1, dim-1)) + [0, dim-1])
                w = weight.reshape(original_shape)
        if fc_spec.HasField('drop_rate'):
            w = w * fc_spec.drop_rate
            if IR_node.attr['use_bias'].b:
                bias = bias * fc_spec.drop_rate


        # weights
        self.set_weight(IR_node.name, 'weights', w)
        if IR_node.attr['use_bias'].b:
            self.set_weight(IR_node.name, 'bias', bias)

        if fc_spec.HasField('active_type') and  fc_spec.active_type != '':
            IR_node_act = self._defuse_activation(source_node)
            PaddleParser._set_output_shape(source_node, IR_node_act, output_shapes)






    def rename_Add(self, source_node):
        add_spec = source_node.layer
        self._convert_merge(source_node, 'Add')
        if add_spec.HasField('active_type') and  add_spec.active_type != '':
            self._defuse_activation(source_node)


    def rename_InputLayer(self, source_node):
        # need the shape TODO

        # only for training
        IR_node = self.IR_graph.node.add()

        # name, op
        PaddleParser._copy_and_reop(source_node, IR_node, "DataInput")

        # input edge
        self.convert_inedge(source_node, IR_node)

        output_shapes = [-1, 224, 224, 3]
        # shape
        PaddleParser._copy_shape(source_node.layer, IR_node, output_shapes)


    def rename_LRN(self, source_node):
        IR_node = self.IR_graph.node.add()

        # name, op
        PaddleParser._copy_and_reop(source_node, IR_node, "LRN")

        # input edge
        self.convert_inedge(source_node, IR_node)

        # layer and spec
        lrn_spec = source_node.layer
        spec = lrn_spec.inputs[0].norm_conf
        channels = spec.channels
        size = spec.size
        alpha = spec.scale
        beta = spec.pow
        img_size_x = spec.img_size
        img_size_y = spec.img_size_y if spec.HasField('img_size_y') else img_size_x
        output_x = spec.output_x
        output_y = spec.output_y if spec.HasField('output_y') else output_x


        # output shape
        output_shapes = [-1, output_y, output_x, channels]
        self.shape_dict[source_node.name] = output_shapes
        PaddleParser._set_output_shape(source_node, IR_node, output_shapes)

        # alpha
        IR_node.attr["alpha"].f = alpha * size
        # beta
        IR_node.attr["beta"].f = beta
        # nsize
        IR_node.attr["size"].i = int((size+1)/2)


        if lrn_spec.HasField('active_type') and  lrn_spec.active_type != '':
            IR_node_act = self._defuse_activation(source_node)
            PaddleParser._set_output_shape(source_node, IR_node_act, output_shapes)





#----------------------------------------------------------------------------------------------
#  Copyright (c) Microsoft Corporation. All rights reserved.
#  Licensed under the MIT License. See License.txt in the project root for license information.
#----------------------------------------------------------------------------------------------

import argparse
import os
from six import text_type as _text_type
import cntk as C
from mmdnn.conversion.common.utils import download_file

BASE_MODEL_URL = 'https://www.cntk.ai/Models/CNTK_Pretrained/'
# pylint: disable=line-too-long
MODEL_URL = {
    'alexnet'              : BASE_MODEL_URL + 'AlexNet_ImageNet_CNTK.model',
    'inception_v3'         : BASE_MODEL_URL + 'InceptionV3_ImageNet_CNTK.model',
    'resnet18'             : BASE_MODEL_URL + 'ResNet18_ImageNet_CNTK.model',
    'resnet50'             : BASE_MODEL_URL + 'ResNet50_ImageNet_CNTK.model',
    'resnet101'            : BASE_MODEL_URL + 'ResNet101_ImageNet_CNTK.model',
    'resnet152'            : BASE_MODEL_URL + 'ResNet152_ImageNet_CNTK.model',
    'Fast-RCNN_grocery100' : 'https://www.cntk.ai/Models/FRCN_Grocery/Fast-RCNN_grocery100.model',
    'Fast-RCNN_Pascal'     : 'https://www.cntk.ai/Models/FRCN_Pascal/Fast-RCNN.model'
}
# pylint: enable=line-too-long


def _main():
    parser = argparse.ArgumentParser()

    parser.add_argument('-n', '--network', type=_text_type, help='Model Type', required=True,
                        choices=MODEL_URL.keys())

    parser.add_argument('-i', '--image', default=None,
                        type=_text_type, help='Test Image Path')

    parser.add_argument('-o', '--output_dir', default='./',
                        type=_text_type, help='CNTK Checkpoint file name')

    args = parser.parse_args()

    fn = download_file(MODEL_URL[args.network], directory=args.output_dir)
    if not fn:
        return -1

    model = C.Function.load(fn)

    if len(model.outputs) > 1:
        for idx, output in enumerate(model.outputs):
            if len(output.shape) > 0:
                eval_node = idx
                break

        model = C.as_composite(model[eval_node].owner)
        model.save(fn)

        print("Model {} is saved as {}.".format(args.network, fn))

    if args.image:
        import numpy as np
        from mmdnn.conversion.examples.imagenet_test import TestKit
        func = TestKit.preprocess_func['cntk'][args.network]
        img = func(args.image)
        img = np.transpose(img, (2, 0, 1))
        predict = model.eval({model.arguments[0]:[img]})
        predict = np.squeeze(predict)
        top_indices = predict.argsort()[-5:][::-1]
        result = [(i, predict[i]) for i in top_indices]
        print(result)
        print(np.sum(result))

    return 0


if __name__ == '__main__':
    _main()



#----------------------------------------------------------------------------------------------
#  Copyright (c) Microsoft Corporation. All rights reserved.
#  Licensed under the MIT License. See License.txt in the project root for license information.
#----------------------------------------------------------------------------------------------

from mmdnn.conversion.common.DataStructure.graph import GraphNode, Graph
from tensorflow.core.framework.node_def_pb2 import NodeDef
from tensorflow.core.framework import attr_value_pb2
import torch


class TorchGraphNode(GraphNode):

    def __init__(self, layer, id):
        # self._type = layer.__class__.__name__.replace('Backward', '')
        # self._name = "{}_{}".format(self.type, id)
        # TODO
        super(PyTorchGraphNode, self).__init__(layer)

    @property
    def name(self):
        return self._name

    @property
    def type(self):
        return self._type


    @property
    def torch_layer(self):
        return self.layer


class TorchGraph(Graph):

    def __init__(self, model):
        super(TorchGraph, self).__init__(model)
        self.model = model


    def build(self, shape):
        print (self.model)
        print (dir(self.model))

        output_shapes = self._infer_torch_output_shapes(
            self.model,
            shape
        )
        print (output_shapes)

        # """
        # build graph for pytorch 0.2.0
        # """
        # dummy_input = torch.autograd.Variable(torch.randn(shape))
        # output_node = self.model(dummy_input)

        # search_queue = [output_node.grad_fn]
        # tmp_node = PyTorchGraphNode(output_node.grad_fn, 0)
        # self.layer_map[tmp_node.name] = tmp_node
        # visited = {output_node.grad_fn : self.layer_map[tmp_node.name]}

        # idx = 0
        # node_count = 1
        # while (idx < len(search_queue)):
        #     current_node = search_queue[idx]
        #     current_type = visited[current_node].type
        #     if hasattr(current_node, 'next_functions'):
        #         for parent, _ in current_node.next_functions:
        #             parent_type = parent.__class__.__name__.replace('Backward', '')
        #             if parent_type != 'AccumulateGrad' and \
        #                (parent_type != 'Transpose' or current_type != 'Addmm'):
        #                 if not parent in visited:
        #                     tmp_node = PyTorchGraphNode(parent, node_count)
        #                     self.layer_map[tmp_node.name] = tmp_node
        #                     node_count += 1
        #                     visited[parent] = tmp_node
        #                     search_queue.append(parent)
        #                 self._make_connection(visited[parent].name, visited[current_node].name)
        #     idx += 1

        super(TorchGraph, self).build()


    @staticmethod
    def _infer_torch_output_shapes(torch_model, input_shapes):
        """
        Forward torch model to infer output shape
        """
        return TorchGraph._forward_torch_random_input(
                torch_model,
                input_shapes,
                is_batch=False)

        # try:
        #     return TorchGraph._forward_torch_random_input(
        #         torch_model,
        #         input_shapes,
        #         is_batch=False
        #     )
        # except:
        #     # try batch mode
        #     # return TorchGraph._forward_torch_random_input(
        #     #     torch_model,
        #     #     input_shapes,
        #     #     is_batch=True
        #     # )
        #     pass

    @staticmethod
    def _forward_torch_random_input(torch_model, input_shapes, is_batch=False):
        input_tensors = []
        for shape in input_shapes:
            if is_batch:
                tensor = torch.rand(1, *shape).float()
            else:
                tensor = torch.randn(shape)
                # tensor = torch.rand(*shape).float()
            input_tensors.append(tensor)

        print (input_tensors[0].shape)
        if len(input_tensors) == 1:
            result = torch_model.forward(input_tensors[0])
        else:
            result = torch_model.forward(input_tensors)

        print ("result", result)
        if isinstance(result, list):
            # multi output
            output_shapes = []
            for tensor in result:
                shape = tensor.numpy().shape
                if is_batch:
                    shape = shape[1:]
                output_shapes.append(shape)
            return output_shapes
        else:
            # single output
            output_shape = result.numpy().shape
            if is_batch:
                return [output_shape[1:]]
            else:
                return [output_shape]


# ----------------------------------------------------------------------------------------------
#  Copyright (c) Microsoft Corporation. All rights reserved.
#  Licensed under the MIT License. See License.txt in the project root for license information.
# ----------------------------------------------------------------------------------------------

from mmdnn.conversion.common.DataStructure.graph import GraphNode, Graph
from onnx import onnx_pb2


class ONNXGraphNode(GraphNode):
    def __init__(self, layer):
        super(ONNXGraphNode, self).__init__(layer)
        self.weights = list()
        self.inputs = list()
        self.outputs = list()

    @property
    def name(self):
        return self.layer.name

    @property
    def type(self):
        return self.layer.op_type

    @property
    def onnx_layer(self):
        return self.layer


# node
#  input
#   edge(node a <-> node b)
#

class ONNXGraph(Graph):
    @staticmethod
    def _generate_name(layer):
        return ""

    def __init__(self, model):
        super(ONNXGraph, self).__init__(model)
        self._graph = model.graph
        # key is edge name, value is src/dst node name
        self._edge_src = dict()
        self._edge_dst = dict()
        # key is initializer name, value is TensorProto
        self._weights = dict()
        self._inputs = dict()
        self._outputs = dict()

    def build(self):
        for w in self._graph.initializer:
            self._weights[w.name] = w
        for s in self._graph.input:
            self._inputs[s.name] = s
        for s in self._graph.output:
            self._outputs[s.name] = s

        for i, layer in enumerate(self._graph.node):
            if not layer.name:
                layer.name = '{0}_{1}'.format(layer.op_type, i)
            name = layer.name
            # print(name)
            # print(layer.op_type)
            node = ONNXGraphNode(layer)
            self.layer_map[name] = node
            self.layer_name_map[name] = name
            for n in layer.input:
                if n in self._weights:
                    # n is input data
                    node.weights.append(n)
                if n in self._inputs:
                    node.inputs.append(n)
                else:
                    # n is input edge
                    self._edge_dst[n] = name
                    if n in self._edge_src:
                        self._make_connection(self._edge_src[n], name)
            for n in layer.output:
                if n in self._outputs:
                    node.outputs.append(n)
                else:
                    self._edge_src[n] = name
                    if n in self._edge_dst:
                        self._make_connection(name, self._edge_dst[n])

        super(ONNXGraph, self).build()



#----------------------------------------------------------------------------------------------
#  Copyright (c) Microsoft Corporation. All rights reserved.
#  Licensed under the MIT License. See License.txt in the project root for license information.
#----------------------------------------------------------------------------------------------

from __future__ import absolute_import
from mmdnn.conversion.examples.imagenet_test import TestKit
from mmdnn.conversion.examples.extractor import base_extractor
from mmdnn.conversion.common.utils import download_file


class mxnet_extractor(base_extractor):

    _base_model_url = 'http://data.mxnet.io/models/'

    _image_size     = 224

    from collections import namedtuple
    Batch = namedtuple('Batch', ['data'])

    architecture_map = {
        'imagenet1k-inception-bn' : {'symbol' : _base_model_url+'imagenet/inception-bn/Inception-BN-symbol.json',
                                     'params' : _base_model_url+'imagenet/inception-bn/Inception-BN-0126.params'},
        'imagenet1k-resnet-18' : {'symbol' : _base_model_url+'imagenet/resnet/18-layers/resnet-18-symbol.json',
                                  'params' : _base_model_url+'imagenet/resnet/18-layers/resnet-18-0000.params'},
        'imagenet1k-resnet-34' : {'symbol' : _base_model_url+'imagenet/resnet/34-layers/resnet-34-symbol.json',
                                  'params' : _base_model_url+'imagenet/resnet/34-layers/resnet-34-0000.params'},
        'imagenet1k-resnet-50' : {'symbol' : _base_model_url+'imagenet/resnet/50-layers/resnet-50-symbol.json',
                                  'params' : _base_model_url+'imagenet/resnet/50-layers/resnet-50-0000.params'},
        'imagenet1k-resnet-101' : {'symbol' : _base_model_url+'imagenet/resnet/101-layers/resnet-101-symbol.json',
                                   'params' : _base_model_url+'imagenet/resnet/101-layers/resnet-101-0000.params'},
        'imagenet1k-resnet-152' : {'symbol' : _base_model_url+'imagenet/resnet/152-layers/resnet-152-symbol.json',
                                   'params' : _base_model_url+'imagenet/resnet/152-layers/resnet-152-0000.params'},
        'imagenet1k-resnext-50' : {'symbol' : _base_model_url+'imagenet/resnext/50-layers/resnext-50-symbol.json',
                                   'params' : _base_model_url+'imagenet/resnext/50-layers/resnext-50-0000.params'},
        'imagenet1k-resnext-101' : {'symbol' : _base_model_url+'imagenet/resnext/101-layers/resnext-101-symbol.json',
                                    'params' : _base_model_url+'imagenet/resnext/101-layers/resnext-101-0000.params'},
        'imagenet1k-resnext-101-64x4d' : {'symbol' : _base_model_url+'imagenet/resnext/101-layers/resnext-101-64x4d-symbol.json',
                                          'params' : _base_model_url+'imagenet/resnext/101-layers/resnext-101-64x4d-0000.params'},
        'imagenet11k-resnet-152' : {'symbol' : _base_model_url+'imagenet-11k/resnet-152/resnet-152-symbol.json',
                                    'params' : _base_model_url+'imagenet-11k/resnet-152/resnet-152-0000.params'},
        'imagenet11k-place365ch-resnet-152' : {'symbol' : _base_model_url+'imagenet-11k-place365-ch/resnet-152-symbol.json',
                                               'params' : _base_model_url+'imagenet-11k-place365-ch/resnet-152-0000.params'},
        'imagenet11k-place365ch-resnet-50' : {'symbol' : _base_model_url+'imagenet-11k-place365-ch/resnet-50-symbol.json',
                                              'params' : _base_model_url+'imagenet-11k-place365-ch/resnet-50-0000.params'},
        'vgg19' : {'symbol' : _base_model_url+'imagenet/vgg/vgg19-symbol.json',
                   'params' : _base_model_url+'imagenet/vgg/vgg19-0000.params'},
        'vgg16' : {'symbol' : _base_model_url+'imagenet/vgg/vgg16-symbol.json',
                   'params' : _base_model_url+'imagenet/vgg/vgg16-0000.params'},
        'squeezenet_v1.0' : {'symbol' : _base_model_url+'imagenet/squeezenet/squeezenet_v1.0-symbol.json',
                             'params' : _base_model_url+'imagenet/squeezenet/squeezenet_v1.0-0000.params'},
        'squeezenet_v1.1' : {'symbol' : _base_model_url+'imagenet/squeezenet/squeezenet_v1.1-symbol.json',
                             'params' : _base_model_url+'imagenet/squeezenet/squeezenet_v1.1-0000.params'}
    }


    @classmethod
    def download(cls, architecture, path="./"):
        if cls.sanity_check(architecture):
            architecture_file = download_file(cls.architecture_map[architecture]['symbol'], directory=path)
            if not architecture_file:
                return None

            weight_file = download_file(cls.architecture_map[architecture]['params'], directory=path)
            if not weight_file:
                return None

            print("MXNet Model {} saved as [{}] and [{}].".format(architecture, architecture_file, weight_file))
            return (architecture_file, weight_file)

        else:
            return None


    @classmethod
    def inference(cls, architecture, files, path, image_path):
        import mxnet as mx
        import numpy as np
        if cls.sanity_check(architecture):
            file_name = cls.architecture_map[architecture]['params'].split('/')[-1]
            prefix, epoch_num = file_name[:-7].rsplit('-', 1)

            sym, arg_params, aux_params = mx.model.load_checkpoint(path + prefix, int(epoch_num))
            model = mx.mod.Module(symbol=sym)
            model.bind(for_training=False,
                       data_shapes=[('data', (1, 3, cls._image_size, cls._image_size))])
            model.set_params(arg_params, aux_params, allow_missing=True, allow_extra=True)

            func = TestKit.preprocess_func['mxnet'][architecture]
            img = func(image_path)
            img = np.transpose(img, [2, 0, 1])
            img = np.expand_dims(img, axis=0)

            model.forward(cls.Batch([mx.nd.array(img)]))
            predict = model.get_outputs()[0].asnumpy()
            predict = np.squeeze(predict)

            del model
            return predict

        else:
            return None



#----------------------------------------------------------------------------------------------
#  Copyright (c) Microsoft Corporation. All rights reserved.
#  Licensed under the MIT License. See License.txt in the project root for license information.
#----------------------------------------------------------------------------------------------

import argparse
import numpy as np
import sys
import os
from mmdnn.conversion.examples.imagenet_test import TestKit
import coremltools

class TestCoreML(TestKit):

    def __init__(self):
        from six import text_type as _text_type
        parser = argparse.ArgumentParser()

        parser.add_argument('-p', '--preprocess', type=_text_type, help='Model Preprocess Type')

        parser.add_argument('--model', '-n', '-w', type=_text_type,
                            required=True, help='CoreML Model path.')

        parser.add_argument('-s', type=_text_type, help='Source Framework Type',
                            choices=self.truth.keys())

        parser.add_argument('--image', '-i',
                            type=_text_type, help='Test image path.',
                            default="mmdnn/conversion/examples/data/seagull.jpg")

        parser.add_argument('-input', type=_text_type,
                            required=True, help='CoreML Input Node')

        parser.add_argument('-output', type=_text_type,
                            required=True, help='CoreML Output Node')

        parser.add_argument('-size', type=int,
            default=224, help='CoreML Input Image Size')


        self.args = parser.parse_args()

        print("Loading model [{}].".format(self.args.model))

        self.model = coremltools.models.MLModel(self.args.model.encode())

        print("Model loading success.")

    def preprocess(self, image_path):
        from PIL import Image as pil_image
        img = pil_image.open(image_path)
        img = img.resize((self.args.size, self.args.size))
        self.data = img

    def print_result(self):
        coreml_inputs = {self.args.input: self.data}
        self.coreml_output = self.model.predict(coreml_inputs, useCPUOnly=False)
        predict = self.coreml_output[self.args.output]
        super(TestCoreML, self).print_result(predict)


    def print_intermediate_result(self, layer_name, if_transpose = False):
        super(TestCoreML, self).print_intermediate_result(self.coreml_output[layer_name], if_transpose)


    def inference(self, image_path):
        self.preprocess(image_path)

        self.print_result()

        # self.print_intermediate_result('conv1_7x7_s2_1', True)

        # self.test_truth()

if __name__=='__main__':
    tester = TestCoreML()
    tester.inference(tester.args.image)



def save_model(MainModel, network_filepath, weight_filepath, dump_filepath):
    model = MainModel.KitModel(weight_filepath)
    model.save(dump_filepath)
    print('Keras model file is saved as [{}], generated by [{}.py] and [{}].'.format(
        dump_filepath, network_filepath, weight_filepath))



import paddle.v2 as paddle

__all__ = ['vgg13', 'vgg16', 'vgg19']


def vgg(input, nums, class_dim):
    def conv_block(input, num_filter, groups, num_channels=None):
        return paddle.networks.img_conv_group(
            input=input,
            num_channels=num_channels,
            pool_size=2,
            pool_stride=2,
            conv_num_filter=[num_filter] * groups,
            conv_filter_size=3,
            conv_act=paddle.activation.Relu(),
            pool_type=paddle.pooling.Max())

    assert len(nums) == 5
    # the channel of input feature is 3
    conv1 = conv_block(input, 64, nums[0], 3)
    conv2 = conv_block(conv1, 128, nums[1])
    conv3 = conv_block(conv2, 256, nums[2])
    conv4 = conv_block(conv3, 512, nums[3])
    conv5 = conv_block(conv4, 512, nums[4])

    fc_dim = 4096
    fc1 = paddle.layer.fc(input=conv5,
                          size=fc_dim,
                          act=paddle.activation.Relu(),
                          layer_attr=paddle.attr.Extra(drop_rate=0.5))
    fc2 = paddle.layer.fc(input=fc1,
                          size=fc_dim,
                          act=paddle.activation.Relu(),
                          layer_attr=paddle.attr.Extra(drop_rate=0.5))
    out = paddle.layer.fc(input=fc2,
                          size=class_dim,
                          act=paddle.activation.Softmax())
    return out


def vgg13(input, class_dim):
    nums = [2, 2, 2, 2, 2]
    return vgg(input, nums, class_dim)


def vgg16(input, class_dim):
    nums = [2, 2, 3, 3, 3]
    return vgg(input, nums, class_dim)


def vgg19(input, class_dim):
    nums = [2, 2, 4, 4, 4]
    return vgg(input, nums, class_dim)



#----------------------------------------------------------------------------------------------
#  Copyright (c) Microsoft Corporation. All rights reserved.
#  Licensed under the MIT License. See License.txt in the project root for license information.
#----------------------------------------------------------------------------------------------

import os
import numpy as np
from six import string_types as _string_types
from mmdnn.conversion.common.IR.IR_graph import IRGraph, IRGraphNode
import mmdnn.conversion.common.IR.graph_pb2 as graph_pb2
from mmdnn.conversion.common.IR.graph_pb2 import NodeDef, GraphDef, DataType
from mmdnn.conversion.common.DataStructure.emitter import Emitter
from mmdnn.conversion.common.utils import *
from mmdnn.conversion.rewriter.folder import Folder

class PytorchEmitter(Emitter):

    dtype_map = {
        graph_pb2.DT_FLOAT16 : "torch.float16",
        graph_pb2.DT_FLOAT32 : "torch.float32",
        graph_pb2.DT_FLOAT64 : "torch.float64",
        graph_pb2.DT_INT16 : "torch.int16",
        graph_pb2.DT_INT32 : "torch.int32",
        graph_pb2.DT_INT64 : "torch.int64",
        graph_pb2.DT_UINT8 : "torch.uint8",
        graph_pb2.DT_UINT16 : "torch.uint16"
    }

    # Base Functions
    def __init__(self, model):
        super(PytorchEmitter, self).__init__()
        if isinstance(model, _string_types):
            network_path = model
        else:
            network_path = model[0]
            weight_path = model[1]

        self.init_code = str()
        self.IR_graph = IRGraph(network_path)
        self.IR_graph.build()
        self._load_weights(weight_path)

        folder = Folder(self.IR_graph, self.weights_dict)
        folder.fold()

    def run(self, dstNetworkPath, dstWeightPath = None, phase = 'test'):
        super(PytorchEmitter, self).run(dstNetworkPath, dstWeightPath, phase)
        if self.weight_loaded:
            self.save_weights(self.weights_dict, dstWeightPath)


    def add_init(self, indent, codes):
        if isinstance(codes, _string_types):
            codes = [codes]
        for code in codes:
            self.init_code += ("    " * indent) + code + '\n'

    def parent_variable_name(self, IR_node, path=[0], weight_type='weights'):
        if not IR_node.in_edges and IR_node.name in self.weights_dict.keys():
            self.weights_dict[IR_node.name][weight_type] = self.weights_dict[IR_node.name][weight_type]
            return "torch.from_numpy(__weights_dict['{}']['{}'])".format(IR_node.name, weight_type)

        return super(PytorchEmitter, self).parent_variable_name(IR_node, path)


    @property
    def header_code(self):
        return """import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F
import math

__weights_dict = dict()

def load_weights(weight_file):
    if weight_file == None:
        return

    try:
        weights_dict = np.load(weight_file).item()
    except:
        weights_dict = np.load(weight_file, encoding='bytes').item()

    return weights_dict

class KitModel(nn.Module):
"""

    def gen_code(self, phase):
        self.add_init(1, """
    def __init__(self, weight_file):
        super(KitModel, self).__init__()
        global __weights_dict
        __weights_dict = load_weights(weight_file)
""")

        self.add_body(1, "def forward(self, x):")

        for layer in self.IR_graph.topological_sort:
            current_node = self.IR_graph.get_node(layer)
            node_type = current_node.type

            if hasattr(self, "emit_" + node_type):
                func = getattr(self, "emit_" + node_type)
                line = func(current_node)
                if line:
                    self.add_body(2, line)

            else:
                print("Pytorch Emitter has not supported operator [%s]." % (node_type))
                self.emit_UNKNOWN(current_node)

        self.add_body(2, "return {}".format(
            ', '.join([self.IR_graph.get_node(name).real_variable_name for name in self.IR_graph.output_layers if self.IR_graph.get_node(name).type != 'Pack'])))

        self.add_body(0, "")
        for i in self.used_layers:
            func = getattr(self, "_layer_" + i)
            func()

        self.add_body(0, "")
        for code in self.layers_codes.values():
            self.add_body(0, code)

        return self.header_code + '\n' + self.init_code + '\n' + self.body_code


    def _defuse_padding(self, IR_node, extra_str = ""):
        input_node = self.parent_variable_name(IR_node)
        if IR_node.get_attr('auto_pad') == 'VALID':
            return input_node

        if is_valid_padding(IR_node.get_attr("pads")) == True:
            return input_node

        padding = self._convert_padding(IR_node)
        input_node = IR_node.variable_name + '_pad'
        self.add_body(2, "{:<15} = F.pad({}, {}{})".format(
            input_node,
            self.parent_variable_name(IR_node),
            padding,
            extra_str
        ))

        return input_node


    def emit_Conv(self, IR_node):
        self.used_layers.add('Conv')

        dim = len(IR_node.get_attr('strides')) - 2

        in_channels = IR_node.get_attr('kernel_shape')[-2]
        filter = IR_node.get_attr('kernel_shape')[-1]
        kernel = IR_node.get_attr('kernel_shape')[:-2]
        strides = IR_node.get_attr('strides')[1:-1]

        if IR_node.type == 'DepthwiseConv':
            group = in_channels
            filter *= group

        else:
            group = IR_node.get_attr('group', 1)

        self.add_init(2, "self.{} = self.__conv({}, name='{}', in_channels={}, out_channels={}, kernel_size={}, stride={}, groups={}, bias={})".format(
            IR_node.variable_name,
            dim,
            IR_node.name,
            in_channels,
            filter,
            tuple(kernel),
            tuple(strides),
            # padding,
            group,
            IR_node.get_attr('use_bias')))

        input_node = self._defuse_padding(IR_node)

        code = "{:<15} = self.{}({})".format(
            IR_node.variable_name,
            IR_node.variable_name,
            input_node)

        if self.weight_loaded:
            if IR_node.type == 'DepthwiseConv':
                self.weights_dict[IR_node.name]['weights'] = np.swapaxes(self.weights_dict[IR_node.name]['weights'], -1, -2)
            self.weights_dict[IR_node.name]['weights'] = np.transpose(self.weights_dict[IR_node.name]['weights'], [dim + 1, dim] + list(range(0, dim)))

        return code


    @staticmethod
    def is_ceil_mode(pads):
        lens = len(pads)
        for i in range(lens // 2 + 1, lens - 1):
            if pads[i] == pads[i - lens // 2]:
                return False
        else:
            return True


    def emit_Pool(self, IR_node):
        dim = len(IR_node.get_attr('strides')) - 2

        if IR_node.get_attr('pooling_type') == "MAX":
            pool_name = "max_pool{}d".format(dim)
            # exstr = ", value=float('-Inf')"
        elif IR_node.get_attr('pooling_type') == "AVG":
            pool_name = "avg_pool{}d".format(dim)
            # exstr = ""
        else:
            raise ValueError()

        if IR_node.layer.attr['global_pooling'].b:
            code = "{:<15} = F.{}(input = {}, kernel_size = {}.size()[2:])".format(
                IR_node.variable_name,
                pool_name,
                self.parent_variable_name(IR_node),
                self.parent_variable_name(IR_node)
            )
            return code

        else:
            if IR_node.get_attr('pooling_type') == "MAX":
                # Change to padding defuse
                input_node = self._defuse_padding(IR_node,", value=float('-inf')")
                for e in IR_node.get_attr('dilations', []):
                    assert e == 1

                pool_size = IR_node.get_attr('kernel_shape')[1:-1]
                strides = IR_node.get_attr('strides')[1:-1]

                code = "{:<15} = F.{}({}, kernel_size={}, stride={}, padding={}, ceil_mode={})".format(
                    IR_node.variable_name,
                    pool_name,
                    input_node,
                    tuple(pool_size),
                    tuple(strides),
                    0,
                    False
                    )
                return code

            elif IR_node.get_attr('pooling_type') == "AVG":

                for e in IR_node.get_attr('dilations', []):
                    assert e == 1

                pool_size = IR_node.get_attr('kernel_shape')[1:-1]
                strides = IR_node.get_attr('strides')[1:-1]

                padding = IR_node.get_attr('pads')[1:dim]
                ceil_mode = self.is_ceil_mode(IR_node.get_attr('pads'))

                # input_node = self._defuse_padding(IR_node, exstr)
                code = "{:<15} = F.{}({}, kernel_size={}, stride={}, padding={}, ceil_mode={})".format(
                    IR_node.variable_name,
                    pool_name,
                    self.parent_variable_name(IR_node),
                    tuple(pool_size),
                    tuple(strides),
                    tuple(padding),
                    ceil_mode
                    )
                return code
            else:
                raise ValueError()


    def emit_UNKNOWN(self, IR_node):
        print(IR_node.name)


    def emit_DataInput(self, IR_node):
        # Ignore it in Pytorch
        IR_node.real_name = 'x'


    def emit_Dropout(self, IR_node):
        code = "{:<15} = F.dropout(input = {}, p = {}, training = self.training, inplace = True)".format(
            IR_node.variable_name,
            self.parent_variable_name(IR_node),
            IR_node.layer.attr["keep_prob"].f)
        return code


    def check_if_need_transpose(self, IR_node):
        parent = self.IR_graph.get_parent(IR_node.name, [0])
        while parent.type == 'Flatten' or parent.type == 'Dropout':
            parent = self.IR_graph.get_parent(parent.name, [0])
        dim = len(parent.layer.attr['_output_shapes'].list.shape[0].dim)
        if dim > 2:
            original_dims = self.weights_dict[IR_node.name]['weights'].shape
            dims = [i.size for i in parent.layer.attr['_output_shapes'].list.shape[0].dim[1:]] + [-1]
            self.weights_dict[IR_node.name]['weights'] = np.reshape(self.weights_dict[IR_node.name]['weights'], dims)
            self.weights_dict[IR_node.name]['weights'] = np.transpose(self.weights_dict[IR_node.name]['weights'], [dim - 2] + list(range(0, dim - 2)) + [dim - 1])
            self.weights_dict[IR_node.name]['weights'] = np.reshape(self.weights_dict[IR_node.name]['weights'], original_dims)


    def emit_FullyConnected(self, IR_node):
        self.used_layers.add(IR_node.type)
        in_features = 1
        for i in self.IR_graph.get_parent(IR_node.name, [0]).layer.attr['_output_shapes'].list.shape[0].dim[1:]:
            in_features *= i.size

        if IR_node.get_attr('in_features') != None:
            in_features = IR_node.get_attr('in_features')

        self.add_init(2, "self.{} = self.__dense(name = '{}', in_features = {}, out_features = {}, bias = {})".format(
            IR_node.variable_name,
            IR_node.name,
            in_features,
            IR_node.layer.attr["units"].i,
            IR_node.IR_layer.attr["use_bias"].b))

        input_node = self.parent_variable_name(IR_node)
        if len(self.IR_graph.get_parent(IR_node.name, [0]).get_attr('_output_shapes')[0].dim) > 2:
            input_node = "{}.view({}.size(0), -1)".format(input_node, input_node)
        
        code = "{:<15} = self.{}({})".format(
            IR_node.variable_name,
            IR_node.variable_name,
            input_node)

        if self.weight_loaded:
            self.check_if_need_transpose(IR_node)
            self.weights_dict[IR_node.name]['weights'] = np.transpose(self.weights_dict[IR_node.name]['weights'], (1, 0))

        return code

    def emit_Flatten(self, IR_node):
        parent = self.IR_graph.get_parent(IR_node.name, [0]).real_variable_name
        code = "{:<15} = {}.view({}.size(0), -1)".format(
            IR_node.variable_name,
            parent,
            parent)
        return code


    def emit_Reshape(self, IR_node):
        shape_list = IR_node.get_attr('shape')
        shape_str = ','.join([str(int(i)) for i in shape_list])
        code = "{:<15} = torch.reshape(input = {}, shape = ({}))".format(
            IR_node.variable_name,
            self.IR_graph.get_node(IR_node.in_edges[0]).real_variable_name,
            shape_str)
        return code


    def emit_Tanh(self, IR_node):
        code = "{:<15} = F.tanh({})".format(
            IR_node.variable_name,
            self.parent_variable_name(IR_node, [0]))
        return code


    def emit_Relu(self, IR_node):
        code = "{:<15} = F.relu({})".format(
            IR_node.variable_name,
            self.parent_variable_name(IR_node, [0]))
        return code


    def emit_LeakyRelu(self, IR_node):
        code = "{:<15} = F.leaky_relu({}, negative_slope={})".format(
            IR_node.variable_name,
            self.parent_variable_name(IR_node, [0]),
            IR_node.get_attr('alpha'))
        return code


    def emit_Relu6(self, IR_node):
        code = "{:<15} = F.relu6({})".format(
            IR_node.variable_name,
            self.parent_variable_name(IR_node, [0]))
        return code


    def emit_Softmax(self, IR_node):
        code = "{:<15} = F.softmax({})".format(
            IR_node.variable_name,
            self.parent_variable_name(IR_node, [0]))
        return code


    def emit_Sigmoid(self, IR_node):
        code = "{:<15} = F.sigmoid({})".format(
            IR_node.variable_name,
            self.parent_variable_name(IR_node)
        )
        return code


    def emit_Embedding(self, IR_node):
        self.used_layers.add("Embedding")
        self.add_init(2, "self.{} = self.__embedding('{}', num_embeddings={}, embedding_dim={})".format(
            IR_node.variable_name,
            IR_node.name,
            IR_node.get_attr('input_dim'),   #2-D
            IR_node.get_attr('output_dim')
            ))
        
        code = "{:<15} = self.{}({})".format(
            IR_node.variable_name,
            IR_node.variable_name,
            "torch.LongTensor(np.array({}))".format(self.parent_variable_name(IR_node))
        )
        return code


    def emit_RNNs(self, IR_node, func):
        raise NotImplementedError()
        # for Keras
        if "dropout" in IR_node.IR_layer.attr:
            dropout_str = ",dropout = {}, recurrent_dropout = {}".format(
                    IR_node.IR_layer.attr['dropout'].f,
                    IR_node.IR_layer.attr['recurrent_dropout'].f)
        else:
            dropout_str = ""

        code = "{:<15} = {}(units = {}, use_bias = {} {})({})".format(
                IR_node.name,
                func,
                IR_node.IR_layer.attr['units'].i,
                IR_node.IR_layer.attr['use_bias'].b,
                dropout_str,
                IR_node.in_edges[0])

        return code


    def emit_LSTM(self, IR_node):
        return self.emit_RNNs(IR_node, "LSTM")


    def emit_GRU(self, IR_node):
        return self.emit_RNNs(IR_node, "GRU")


    def emit_Add(self, IR_node):
        code = "{:<15} = {} + {}".format(
            IR_node.variable_name,
             self.parent_variable_name(IR_node),
             self.parent_variable_name(IR_node, [1]))
        return code


    def emit_Sub(self, IR_node):
        code = "{:<15} = {}".format(
            IR_node.variable_name,
            ' - '.join(self.parent_variable_name(IR_node, [idx]) for idx in range(len(IR_node.in_edges))))
        return code


    def emit_Mul(self, IR_node):
        code = "{:<15} = {}".format(
            IR_node.variable_name,
            ' * '.join(self.parent_variable_name(IR_node, [idx]) for idx in range(len(IR_node.in_edges))))
        return code


    def emit_MatMul(self, IR_node):
        code = "{:<15} = torch.matmul({})".format(
            IR_node.variable_name,
            ' , '.join('%s' % self.IR_graph.get_node(s).real_variable_name for s in IR_node.in_edges))
        return code


    def emit_Constant(self, IR_node):
        if IR_node.get_attr('value'):
            value = IR_node.get_attr('value')
            if not isinstance(value, list):
                value = [value]
            code = "self.{:<15} = torch.autograd.Variable(torch.Tensor({}), requires_grad=False)".format(
                IR_node.variable_name,
                value)
        else:
            code = "self.{:<15} = torch.autograd.Variable(torch.from_numpy(__weights_dict['{}']['value']), requires_grad=False)".format(
                IR_node.variable_name,
                IR_node.name)
        
        # self.add_init(2, "self.{:<15} = torch.from_numpy(__weights_dict['{}']['value'])".format(
        #     IR_node.variable_name,
        #     IR_node.name))
        IR_node.real_name = "self." + IR_node.variable_name
        return code


    def _convert_axis(self, IR_node, axis):
        ndim = len(self.IR_graph.get_parent(IR_node.name, [0]).get_attr('_output_shapes')[0].dim)
        if axis == 0:
            return 0
        elif axis == ndim - 1:
            return 1
        else:
            return axis + 1


    def emit_Concat(self, IR_node):
        axis = self._convert_axis(IR_node, IR_node.get_attr('axis'))
        code = "{:<15} = torch.cat(({}), {})".format(
            IR_node.variable_name,
            ', '.join(self.parent_variable_name(IR_node, [idx]) for idx in range(len(IR_node.in_edges))),
            axis,
        )
        return code


    def emit_BatchNorm(self, IR_node):
        self.used_layers.add(IR_node.type)
        dim = len(IR_node.layer.attr['_output_shapes'].list.shape[0].dim) - 2

        output_shape = IR_node.layer.attr['_output_shapes'].list.shape[0]
        if IR_node.get_attr('data_format', "NHWC") == "NCHW":
            num_features = output_shape.dim[1].size
        else:
            num_features = output_shape.dim[-1].size

        self.add_init(2, "self.{} = self.__batch_normalization({}, '{}', num_features={}, eps={}, momentum={})".format(
             IR_node.variable_name,
             dim,
             IR_node.name,
             num_features,
             IR_node.layer.attr['epsilon'].f,
             IR_node.layer.attr['momentum'].f,
        ))

        code = "{:<15} = self.{}({})".format(
            IR_node.variable_name,
            IR_node.variable_name,
            self.parent_variable_name(IR_node)
        )
        return code


    def emit_Scale(self, IR_node):
        self.used_layers.add(IR_node.type)
        dim = len(IR_node.layer.attr['_output_shapes'].list.shape[0].dim) - 2

        self.add_init(2, "self.{} = self.__scale({}, '{}', num_features={})".format(
             IR_node.variable_name,
             dim,
             IR_node.name,
             IR_node.layer.attr['_output_shapes'].list.shape[0].dim[-1].size
        ))

        code = "{:<15} = self.{}({})".format(
            IR_node.variable_name,
            IR_node.variable_name,
            self.parent_variable_name(IR_node)
        )
        return code


    def emit_Squeeze(self, IR_node):
        code = "{:<15} = torch.squeeze({})".format(
            IR_node.variable_name, self.parent_variable_name(IR_node)
        )
        return code


    @staticmethod
    def _convert_padding(IR_node):
        padding = IR_node.get_attr('pads')
        padding = convert_onnx_pad_to_tf(padding)[1:-1]
        new_padding = []
        for pad in padding:
            new_padding.insert(0, pad)
        return tuple(np.array(new_padding).reshape(-1).tolist())


    def emit_Pad(self, IR_node):
        if IR_node.get_attr('mode').lower() == 'constant':
            mode = "mode = 'constant', value = {}".format(0)
        elif IR_node.get_attr('mode').lower() == 'reflect':
            mode = "mode = 'reflect'"
        elif IR_node.get_attr('mode').upper() == 'SYMMETRIC':
            mode = "mode = 'replicate'"
        else:
            assert False

        padding = self._convert_padding(IR_node)
        code = "{:<15} = F.pad({}, {}, {})".format(
            IR_node.variable_name,
            self.parent_variable_name(IR_node),
            padding,
            mode)
        return code


    def emit_ReduceMean(self, IR_node):
        axes = [self._convert_axis(IR_node, x) for x in IR_node.get_attr('axes')]
        input_node = self.parent_variable_name(IR_node)
        codes = []
        for axis in sorted(axes, reverse=True):
            code = "{:<15} = torch.mean({}, {}, {})".format(
                IR_node.variable_name,
                input_node,
                axis,
                IR_node.get_attr("keepdims")
            )
            codes.append(code)
            input_node = IR_node.variable_name
        return codes


    def emit_LRN(self, IR_node):
        code =  "{:<15} = F.local_response_norm({}, size={}, alpha={}, beta={}, k={})".format(
            IR_node.variable_name,
            self.parent_variable_name(IR_node),
            IR_node.get_attr('size') * 2 - 1,
            IR_node.get_attr('alpha'),
            IR_node.get_attr('beta'),
            IR_node.get_attr('k', 1)
        )
        return code


    def emit_DepthwiseConv(self, IR_node):
        return self.emit_Conv(IR_node)


    def emit_Const(self, IR_node):
        if 'dtype' in IR_node.layer.attr:
            dtype_str = "dtype={}".format(self.dtype_map[IR_node.layer.attr['dtype'].type])
            if 'int' in dtype_str:
                code = "{:<15} = torch.tensor({}, {})".format(
                    IR_node.variable_name,
                    IR_node.layer.attr['value'].i,
                    dtype_str)
            else:
                code = "{:<15} = torch.tensor({}, {})".format(
                    IR_node.variable_name,
                    IR_node.layer.attr['value'].f,
                    dtype_str)

        else:
            dtype_str = "dtype=torch.float32"
            code = "{:<15} = torch.tensor({}, {})".format(
                IR_node.variable_name,
                IR_node.layer.attr['value'].f,
                dtype_str)
        return code


    def emit_Shape(self, IR_node):
        code = "{:<15} = torch.Tensor(list({}.size()))".format(
            IR_node.variable_name,
            self.parent_variable_name(IR_node)
            )
        return code


    def emit_Pack(self, IR_node):
        code = "{:<15} = {}".format(
            IR_node.variable_name,
            '[' +  ','.join('%s' % self.IR_graph.get_node(s).real_variable_name for s in IR_node.in_edges) + ']',
            )
        return code


    def emit_Slice(self, IR_node):
        starts = IR_node.get_attr('starts')
        if len(starts) > 1:
            starts = [starts[0], starts[-1]] + starts[1:-1]
        ends = IR_node.get_attr('ends')
        if len(ends) > 1:
            ends = [ends[0], ends[-1]] + ends[1:-1]
        extra_str = ""
        for idx, _ in enumerate(starts):
            if idx:
                extra_str += ", "
            extra_str += "{}:".format(starts[idx])
            if ends[idx]:
                extra_str += "{}".format(ends[idx])
        
        shrink_mask = IR_node.get_attr('shrink_axis_mask')

        if shrink_mask:
            mask = [int(s) for s in bin(shrink_mask)[2:][::-1]]
            shrink_str = '[' + ','.join(':' if bit==0 else '0' for bit in mask) + ']'
        else:
            shrink_str = ''
        code = "{:<15} = {}[{}]{}".format(
            IR_node.variable_name,
            self.parent_variable_name(IR_node),
            extra_str,
            shrink_str
        )
        return code


    def emit_Split(self, IR_node):

        if isinstance(IR_node.get_attr('split'), list):
            split_str = IR_node.get_attr('split')
        else:
            num_split = IR_node.get_attr('split')
            split_str = "math.ceil({}.shape[{}]/{})".format(
                self.parent_variable_name(IR_node), 
                IR_node.get_attr('axis'),
                num_split)
        code = "{:<15} = torch.split({}, {}, dim={})".format(
            IR_node.variable_name,
            self.parent_variable_name(IR_node),
            split_str,
            IR_node.get_attr('axis'),
        )
        return code


    def emit_Unstack(self, IR_node):
        code = "{:<15} = torch.unbind({}, dim={})".format(
            IR_node.variable_name,
            self.parent_variable_name(IR_node),
            IR_node.get_attr('axis')
        )
        return code


    def emit_Fill(self, IR_node):
        code = "{:<15} = torch.full({}.int().numpy().tolist(), {})".format(
            IR_node.variable_name,
            self.parent_variable_name(IR_node),
            IR_node.get_attr('value')
        )
        return code


    def emit_Gather(self, IR_node):
        pass


    def emit_Unsqueeze(self, IR_node):
        code = "{:<15} = {}.unsqueeze({})".format(
            IR_node.variable_name,
            self.parent_variable_name(IR_node),
            IR_node.get_attr('axes')[0]
        )
        return code


    def emit_Transpose(self, IR_node):
        code = "{:<15} = {}.permute({})".format(
            IR_node.variable_name,
            self.parent_variable_name(IR_node),
            self.parent_variable_name(IR_node, [1]))
        return code


    def emit_Minimum(self, IR_node):
        code = "{:<15} = torch.min({}, {})".format(
            IR_node.variable_name,
            self.parent_variable_name(IR_node),
            self.parent_variable_name(IR_node, [1]))
        return code


    def emit_Maxmum(self, IR_node):
        code = "{:<15} = torch.max({}, {})".format(
            IR_node.variable_name,
            self.parent_variable_name(IR_node),
            self.parent_variable_name(IR_node, [1]))
        return code


    def emit_Scope(self, IR_node):
        input_vars = [self.parent_variable_name(IR_node, [idx]) for idx in range(len(IR_node.in_edges))]
        code = "{:<15} = self.__{}({})".format(
            IR_node.real_variable_name,
            IR_node.pattern,
            ', '.join(input_vars))
        self._gen_scope_code(IR_node)
        return code


    def _gen_scope_code(self, scope_node):

        def _scope_func(scope_name, params, code, return_var):
            code = """
    def __{}({}):
{}
        return {}
    """.format(scope_name, params, code, ', '.join(return_var))
            return code

        if not self.layers_codes.get(scope_node.pattern, None):
            body_code = str()
            for node_name in scope_node.topology_list:
                node = self.IR_graph.get_node(node_name)
                node_type = node.type

                if hasattr(self, "emit_" + node_type):
                    func = getattr(self, "emit_" + node_type)
                    line = func(node)
                    if line != None:
                        body_code += "        " + line + '\n'
                else:
                    print("PytorchEmitter has not supported operator [%s]." % (node_type))
                    self.emit_UNKNOWN(node)

            # param_code does not need parameter slice.
            input_params = scope_node.input_params
            input_params.insert(0, "self")
            param_code = ', '.join(input_params)
            function_code = _scope_func(scope_node.pattern, param_code, body_code, scope_node.return_variables)

            self.layers_codes[scope_node.pattern] = function_code


    def _layer_Embedding(self):
        self.add_body(0,"""
    @staticmethod
    def __embedding(name, **kwargs):
        layer = nn.Embedding(**kwargs) #shape
        layer.state_dict()['weight'].copy_(torch.from_numpy(__weights_dict[name]['weights']))
        return layer
        """)


    def _layer_Conv(self):
        self.add_body(0, """
    @staticmethod
    def __conv(dim, name, **kwargs):
        if   dim == 1:  layer = nn.Conv1d(**kwargs)
        elif dim == 2:  layer = nn.Conv2d(**kwargs)
        elif dim == 3:  layer = nn.Conv3d(**kwargs)
        else:           raise NotImplementedError()

        layer.state_dict()['weight'].copy_(torch.from_numpy(__weights_dict[name]['weights']))
        if 'bias' in __weights_dict[name]:
            layer.state_dict()['bias'].copy_(torch.from_numpy(__weights_dict[name]['bias']))
        return layer""")


    def _layer_FullyConnected(self):
        self.add_body(0, """
    @staticmethod
    def __dense(name, **kwargs):
        layer = nn.Linear(**kwargs)
        layer.state_dict()['weight'].copy_(torch.from_numpy(__weights_dict[name]['weights']))
        if 'bias' in __weights_dict[name]:
            layer.state_dict()['bias'].copy_(torch.from_numpy(__weights_dict[name]['bias']))
        return layer""")


    def _layer_BatchNorm(self):
        self.add_body(0, """
    @staticmethod
    def __batch_normalization(dim, name, **kwargs):
        if   dim == 1:  layer = nn.BatchNorm1d(**kwargs)
        elif dim == 2:  layer = nn.BatchNorm2d(**kwargs)
        elif dim == 3:  layer = nn.BatchNorm3d(**kwargs)
        else:           raise NotImplementedError()

        if 'scale' in __weights_dict[name]:
            layer.state_dict()['weight'].copy_(torch.from_numpy(__weights_dict[name]['scale']))
        else:
            layer.weight.data.fill_(1)

        if 'bias' in __weights_dict[name]:
            layer.state_dict()['bias'].copy_(torch.from_numpy(__weights_dict[name]['bias']))
        else:
            layer.bias.data.fill_(0)

        layer.state_dict()['running_mean'].copy_(torch.from_numpy(__weights_dict[name]['mean']))
        layer.state_dict()['running_var'].copy_(torch.from_numpy(__weights_dict[name]['var']))
        return layer""")


    def _layer_Scale(self):
        self.add_body(0, """
    # from torch.nn.parameter import Parameter

    class _Scale(nn.Module):

        def __init__(self, num_features, affine=True):
            super(KitModel._Scale, self).__init__()
            self.num_features = num_features
            self.affine = affine

            self.running_mean = torch.zeros(num_features)
            self.running_var = torch.ones(num_features)
            self.training = False
            self.eps = 1e-5
            if self.affine:
                self.weight = nn.Parameter(torch.Tensor(num_features))
                self.bias = nn.Parameter(torch.Tensor(num_features))
            else:
                self.register_parameter('weight', None)
                self.register_parameter('bias', None)
            self.reset_parameters()


        def reset_parameters(self):
            if self.affine:
                self.weight.data.uniform_()
                self.bias.data.zero_()

        def _check_input_dim(self, input):
            raise NotImplementedError

        def forward(self, input):
            self._check_input_dim(input)

            return F.batch_norm(
                input, self.running_mean, self.running_var, self.weight, self.bias,
                self.training,
                0 , self.eps)


    class Scale1d(_Scale):

        def _check_input_dim(self, input):
            if input.dim() != 2 and input.dim() != 3:
                raise ValueError('expected 2D or 3D input (got {}D input)'
                                .format(input.dim()))



    class Scale2d(_Scale):


        def _check_input_dim(self, input):
            if input.dim() != 4:
                raise ValueError('expected 4D input (got {}D input)'
                                .format(input.dim()))


    class Scale3d(_Scale):

        def _check_input_dim(self, input):
            if input.dim() != 5:
                raise ValueError('expected 5D input (got {}D input)'
                                .format(input.dim()))


    @staticmethod
    def __scale(dim, name, **kwargs):
        if   dim == 1:  layer = KitModel.Scale1d(**kwargs)
        elif dim == 2:  layer = KitModel.Scale2d(**kwargs)
        elif dim == 3:  layer = KitModel.Scale3d(**kwargs)
        else:           raise NotImplementedError()

        if 'scale' in __weights_dict[name]:
            layer.state_dict()['weight'].copy_(torch.from_numpy(__weights_dict[name]['scale']))
        else:
            layer.weight.data.fill_(1)

        if 'bias' in __weights_dict[name]:
            layer.state_dict()['bias'].copy_(torch.from_numpy(__weights_dict[name]['bias']))
        else:
            layer.bias.data.fill_(0)

        return layer""")



#----------------------------------------------------------------------------------------------
#  Copyright (c) Microsoft Corporation. All rights reserved.
#  Licensed under the MIT License. See License.txt in the project root for license information.
#----------------------------------------------------------------------------------------------

from __future__ import division

import os
import sys
import math
import numpy as np

import caffe
from caffe import layers as L
from caffe import params as P
from mmdnn.conversion.common.IR.IR_graph import IRGraph, IRGraphNode
import mmdnn.conversion.common.IR.graph_pb2 as graph_pb2
from mmdnn.conversion.common.IR.graph_pb2 import NodeDef, GraphDef, DataType
from mmdnn.conversion.common.DataStructure.emitter import Emitter
from mmdnn.conversion.common.utils import *


class CaffeEmitter(Emitter):

    def __init__(self, model):
        from six import string_types as _string_types
        super(CaffeEmitter, self).__init__()
        if isinstance(model, _string_types):
            network_path = model
        else:
            network_path = model[0]
            self._load_weights(model[1])

        self.IR_graph = IRGraph(network_path)
        super(CaffeEmitter, self)._build()


    @property
    def header_code(self):
        return """from __future__ import print_function
import numpy as np
import sys, argparse
import caffe
from caffe import layers as L
from caffe import params as P
from caffe import to_proto
from six import text_type as _text_type


__weights_dict = dict()

def load_weights(weight_file):
    if weight_file == None:
        return

    try:
        weights_dict = np.load(weight_file).item()
    except:
        weights_dict = np.load(weight_file, encoding='bytes').item()

    return weights_dict


def KitModel(weight_file = None):
    n = caffe.NetSpec()
"""

    @property
    def end_code(self):
        return """    return n

def make_net(prototxt):
    n = KitModel()
    with open(prototxt, 'w') as fpb:
        print(n.to_proto(), file=fpb)

def gen_weight(weight_file, model, prototxt):
    global __weights_dict
    __weights_dict = load_weights(weight_file)

    net = caffe.Net(prototxt, caffe.TRAIN)

    for key in __weights_dict:
        if 'weights' in __weights_dict[key]:
            net.params[key][0].data.flat = __weights_dict[key]['weights']
        elif 'mean' in __weights_dict[key]:
            net.params[key][0].data.flat = __weights_dict[key]['mean']
            net.params[key][1].data.flat = __weights_dict[key]['var']
            if 'scale' in __weights_dict[key]:
                net.params[key][2].data.flat = __weights_dict[key]['scale']
        elif 'scale' in __weights_dict[key]:
            net.params[key][0].data.flat = __weights_dict[key]['scale']
        if 'bias' in __weights_dict[key]:
            net.params[key][1].data.flat = __weights_dict[key]['bias']
        if 'gamma' in __weights_dict[key]: # used for prelu, not sure if other layers use this too
            net.params[key][0].data.flat = __weights_dict[key]['gamma']
    net.save(model)
    return net



if __name__=='__main__':
    parser = argparse.ArgumentParser(description='Generate caffe model and prototxt')
    parser.add_argument('--weight_file', '-w', type=_text_type, default='IR weight file')
    parser.add_argument('--prototxt', '-p', type=_text_type, default='caffe_converted.prototxt')
    parser.add_argument('--model', '-m', type=_text_type, default='caffe_converted.caffemodel')
    args = parser.parse_args()
    # For some reason argparser gives us unicode, so we need to conver to str first
    make_net(str(args.prototxt))
    gen_weight(str(args.weight_file), str(args.model), str(args.prototxt))

"""

    def gen_code(self, phase = 'test'):
        self.phase = phase
        self.add_body(0, self.header_code)

        #for test
        # with open("graph.txt", 'w') as f:
        #     for layer in self.IR_graph.topological_sort:
        #         current_node = self.IR_graph.get_node(layer)
        #         print("========current_node=========\n{}".format(current_node.layer), file=f)
        #test end

        for layer in self.IR_graph.topological_sort:
            current_node = self.IR_graph.get_node(layer)
            node_type = current_node.type
            #print("========current_node={}".format(current_node.layer))

            if hasattr(self, "emit_" + node_type):
                func = getattr(self, "emit_" + node_type)
                func(current_node)
            else:
                print("CaffeEmitter has not supported operator [%s]." % (node_type))
                self.emit_UNKNOWN(current_node)

        self.add_body(0, "")
        self.add_body(0,self.end_code)

        return self.body_code


    def run(self, dstNetworkPath, dstWeightPath = None, phase = 'test'):
        super(CaffeEmitter, self).run(dstNetworkPath, dstWeightPath, phase)
        if self.weight_loaded:
            self.save_weights(self.weights_dict, dstWeightPath)


    @staticmethod
    def _shapeToStr(shapes):
        return [dim.size if dim.size > 0 else 1 for dim in shapes.dim]


    def _get_symmetric_padding(self, IR_node):
        stride_h = IR_node.get_attr('strides')[1]
        stride_w = IR_node.get_attr('strides')[2]

        # check if have pad layer
        IR_parent_node = self.IR_graph.get_parent(IR_node.name, [0])
        if IR_parent_node.type == 'Pad':
            pads = IR_parent_node.get_attr('pads')
        else:
            pads = IR_node.get_attr('pads')

        # Pad_h < kernel_h (vgg19 caffe2caffe)
        if IR_node.type == "Pool":
            if pads[1]:
                pad_h = pads[1] + (0 if pads[1] == pads[5] else stride_h)
            else:
                pad_h = 0
            if pads[2]:
                pad_w = pads[2] + (0 if pads[2] == pads[6] else stride_w)
            else:
                pad_w = 0
        else:
            pad_h = pads[1] + (0 if pads[1] == pads[5] else stride_h)
            pad_w = pads[2] + (0 if pads[2] == pads[6] else stride_w)

        return pad_h, pad_w


    def check_if_need_transpose(self, IR_node):
        parent = self.IR_graph.get_parent(IR_node.name, [0])
        while parent.type == 'Flatten' or parent.type == 'Dropout' or parent.type == 'Reshape':
            parent = self.IR_graph.get_parent(parent.name, [0])
        dim = len(parent.layer.attr['_output_shapes'].list.shape[0].dim)
        if dim > 2:
            original_dims = self.weights_dict[IR_node.name]['weights'].shape
            dims = [i.size for i in parent.layer.attr['_output_shapes'].list.shape[0].dim[1:]] + [-1]
            self.weights_dict[IR_node.name]['weights'] = np.reshape(self.weights_dict[IR_node.name]['weights'], dims)
            self.weights_dict[IR_node.name]['weights'] = np.transpose(self.weights_dict[IR_node.name]['weights'], [dim - 2] + list(range(0, dim - 2)) + [dim - 1])
            self.weights_dict[IR_node.name]['weights'] = np.reshape(self.weights_dict[IR_node.name]['weights'], original_dims)


    def emit_Conv(self, IR_node):
        # implement asymmetric paddings by applying symmetric padding then cropping
        pad_h, pad_w = self._get_symmetric_padding(IR_node)

        num_output = IR_node.get_attr('kernel_shape')[-1]
        if IR_node.type == "DepthwiseConv":
            num_group = IR_node.get_attr("kernel_shape")[-2]
            num_output = IR_node.get_attr('kernel_shape')[-2]
        else:
            num_group = IR_node.get_attr("group", 1)

        self.add_body(1, "n.{:<15} = L.Convolution(n.{}, kernel_h={}, kernel_w={}, stride={}, num_output={}, pad_h={}, pad_w={}, group={}, \
            bias_term={}, ntop=1)".format(
            IR_node.variable_name,
            self.parent_variable_name(IR_node),
            IR_node.get_attr('kernel_shape')[0],
            IR_node.get_attr('kernel_shape')[1],
            IR_node.get_attr('strides')[1],
            num_output,
            pad_h,
            pad_w,
            num_group,
            IR_node.get_attr('use_bias', False)))

        dim = len(IR_node.get_attr('strides')) - 2
        if self.weight_loaded:
            if IR_node.type == "DepthwiseConv":
                self.weights_dict[IR_node.name]['weights'] = np.swapaxes(self.weights_dict[IR_node.name]['weights'], -1, -2)
            self.weights_dict[IR_node.name]['weights'] = np.transpose(self.weights_dict[IR_node.name]['weights'], [dim + 1, dim] + list(range(0, dim)))
            self.weights_dict[IR_node.variable_name] = self.weights_dict.pop(IR_node.name)

        self.check_if_need_crop(IR_node)
        # keys = []
        # for key in self.weights_dict[IR_node.name].keys():
        #     keys.append(key)
        # print("=======Layer: {}, keys: {}".format(IR_node.name, keys))

    def compute_output_shape(self, IR_node, kernel_h, kernel_w):
        parent_node = self.IR_graph.get_parent(IR_node.name, [0])

        if parent_node.get_attr('_output_shapes'):
            shape = parent_node.get_attr('_output_shapes')[0]
            shape = shape_to_list(shape)
            h_i = shape[1]
            w_i = shape[2]
            pad_h, pad_w = self._get_symmetric_padding(IR_node)
            stride_h = IR_node.get_attr('strides')[1]
            stride_w = IR_node.get_attr('strides')[2]

            if IR_node.type == 'Pool':
                h_o = (h_i + 2 * pad_h - kernel_h + stride_h - 1) // stride_h + 1
                w_o = (w_i + 2 * pad_w - kernel_w + stride_w - 1) // stride_w + 1
            else:
                h_o = (h_i + 2 * pad_h - kernel_h) // stride_h + 1
                w_o = (w_i + 2 * pad_w - kernel_w) // stride_w + 1
            return h_o, w_o
        else:
            assert False


    def check_if_need_crop(self, IR_node):
        shape = IR_node.get_attr('_output_shapes')[0]
        shape = shape_to_list(shape)
        ir_ho = shape[1]
        ir_wo = shape[2]
        if ir_ho <0 or ir_wo<0:
            return
        if IR_node.type == 'Pool':
            k_h = IR_node.get_attr('kernel_shape')[1]
            k_w = IR_node.get_attr('kernel_shape')[2]
        else:
            k_h = IR_node.get_attr('kernel_shape')[0]
            k_w = IR_node.get_attr('kernel_shape')[1]

        caffe_ho, caffe_wo = self.compute_output_shape(IR_node, k_h, k_w)

        # if asymmetric padding, set offset to 1
        pads = IR_node.get_attr('pads')
        offset = [0 if pads[1] == pads[5] else 1,
                  0 if pads[2] == pads[6] else 1]
        if caffe_ho > ir_ho or caffe_wo > ir_wo:
            crop_layer_variable_name = IR_node.variable_name + "_crop"
            self.add_body(1, "n.{:<15} = L.Crop(n.{}, L.DummyData(shape=[dict(dim=[1, {}, {}, {}])], \
                ntop=1), ntop=1, offset={})".format(
                crop_layer_variable_name,
                IR_node.variable_name,
                shape[3],
                ir_ho,
                ir_wo,
                offset
            ))
            # Change the layer name
            IR_node.real_name = IR_node.real_name + "_crop"


    def emit_Pool(self, IR_node):
        pooling_type = IR_node.get_attr('pooling_type')
        if pooling_type == 'MAX':
            pooling_type = P.Pooling.MAX
        elif pooling_type == 'AVG':
            pooling_type = P.Pooling.AVE
        elif pooling_type == 'STOCHASTIC':
            pooling_type = P.Pooling.STOCHASTIC
        else:
            raise ValueError()

        if IR_node.layer.attr['global_pooling'].b:
            self.add_body(1, "n.{:<15} = L.Pooling(n.{}, pool={}, stride={}, global_pooling=True, ntop=1)".format(
                IR_node.variable_name,
                self.parent_variable_name(IR_node),
                pooling_type,
                IR_node.get_attr('strides')[1]))
        else:
            pad_h, pad_w = self._get_symmetric_padding(IR_node)
            pool_size = IR_node.get_attr('kernel_shape')[1:3]
            if pool_size[0] != pool_size[1]:
                self.add_body(1, "n.{:<15} = L.Pooling(n.{}, pool={}, kernel_h={}, kernel_w={}, pad_h={}, pad_w={}, stride={}, ntop=1)".format(
                    IR_node.variable_name,
                    self.parent_variable_name(IR_node),
                    pooling_type,
                    pool_size[0],
                    pool_size[1],
                    pad_h,
                    pad_w,
                    IR_node.get_attr('strides')[1]))
            else:
                self.add_body(1, "n.{:<15} = L.Pooling(n.{}, pool={}, kernel_size={}, pad_h={}, pad_w={}, stride={}, ntop=1)".format(
                    IR_node.variable_name,
                    self.parent_variable_name(IR_node),
                    pooling_type,
                    pool_size[0],
                    pad_h,
                    pad_w,
                    IR_node.get_attr('strides')[1]))

            # check if need crop output shape
            self.check_if_need_crop(IR_node)

    def emit_ResizeBilinear(self, IR_node):
        shape = IR_node.get_attr("_output_shapes")[0]
        shape = shape_to_list(shape)
        self.add_body(1, "n.{:<15} = L.ResizeBilinear(n.{}, height={}, width={}, ntop=1)".format(
            IR_node.variable_name,
            self.parent_variable_name(IR_node),
            shape[1],
            shape[2]))

    def emit_UNKNOWN(self, IR_node):
        print(IR_node.IR_layer.name)


    def emit_DataInput(self, IR_node):
        shape = self._shapeToStr(IR_node.get_attr('shape'))
        shape = [shape[0], shape[-1]] + shape[1:-1]
        self.add_body(1, "n.{:<15} = L.Input(shape=[dict(dim={})], ntop=1)".format(
            IR_node.variable_name,
            shape))


    def emit_Dropout(self, IR_node):
        in_place = True
        self.add_body(1, "n.{:<15} = L.Dropout(n.{}, dropout_ratio={} , in_place={}, ntop=1)".format(
            IR_node.variable_name,
            self.parent_variable_name(IR_node),
            1 - IR_node.get_attr('keep_prob'),
            in_place))


    def emit_FullyConnected(self, IR_node):
        self.add_body(1, "n.{:<15} = L.InnerProduct(n.{}, num_output={}, bias_term={}, ntop=1)".format(
            IR_node.variable_name,
            self.parent_variable_name(IR_node),
            IR_node.layer.attr["units"].i,
            IR_node.get_attr('use_bias', False)))
        if self.weight_loaded:
            self.check_if_need_transpose(IR_node)
            self.weights_dict[IR_node.name]['weights'] = np.transpose(self.weights_dict[IR_node.name]['weights'], (1, 0))
            self.weights_dict[IR_node.variable_name] = self.weights_dict.pop(IR_node.name)


    def emit_BatchNorm(self, IR_node):

        self.add_body(1, "n.{:<15} = L.BatchNorm(n.{}, eps={}, use_global_stats={}, ntop=1)".format(
            IR_node.variable_name,
            self.parent_variable_name(IR_node),
            IR_node.get_attr('epsilon'),
            self.phase == 'test'
        ))

        scale_layer_var_name = IR_node.variable_name + "_scale"
        self.add_body(1, "n.{:<15} = L.Scale(n.{}, bias_term={}, in_place=True, ntop=1)".format(
            scale_layer_var_name,
            IR_node.variable_name,
            IR_node.get_attr('bias', False)
        ))

        if self.weight_loaded:
            self.weights_dict[scale_layer_var_name] = dict()
            if 'scale' in self.weights_dict[IR_node.name]:
                self.weights_dict[scale_layer_var_name]['scale'] = self.weights_dict[IR_node.name]['scale']
            else:
                self.weights_dict[scale_layer_var_name]['scale'] = 1

            self.weights_dict[IR_node.name]['scale'] = 1

            if 'bias' in self.weights_dict[IR_node.name]:
                self.weights_dict[scale_layer_var_name]['bias'] = self.weights_dict[IR_node.name]['bias']
                self.weights_dict[IR_node.name].pop('bias', None)
                # change the key "name" to "variable_name", in case of the layer name has invalid characters

            self.weights_dict[IR_node.variable_name] = self.weights_dict.pop(IR_node.name)

        IR_node.real_name = IR_node.name + "_scale"


    def emit_Scale(self, IR_node):
        self.add_body(1, "n.{:<15} = L.Scale(n.{}, bias_term={}, in_place=True, ntop=1)".format(
            IR_node.variable_name,
            self.parent_variable_name(IR_node),
            IR_node.get_attr('use_bias', False)
        ))
        if self.weight_loaded:
            self.weights_dict[IR_node.variable_name] = self.weights_dict.pop(IR_node.name)


    def emit_Constant(self, IR_node):
        if IR_node.get_attr('value'):
            value = IR_node.get_attr('value')
        else:
            value = self.weights_dict[IR_node.name]['value'][0]
        IR_node_after = self.IR_graph.get_son(IR_node.name, [0])
        shape = IR_node_after.get_attr("_output_shapes")[0]
        shape = shape_to_list(shape)
        self.add_body(1, "n.{:<15} = L.DummyData(shape=[dict(dim=[1,{},{},{}])], data_filler=dict(type='constant', value={}), ntop=1)".format(
            IR_node.variable_name,
            shape[-1],
            shape[1],
            shape[2],
            value
        ))


    def emit_LRN(self, IR_node):
        self.add_body(1, "n.{:<15} = L.LRN(n.{}, local_size={}, alpha={}, beta={}, k={})".format(
            IR_node.variable_name,
            self.parent_variable_name(IR_node),
            IR_node.get_attr('size') * 2 - 1,
            IR_node.get_attr('alpha'),
            IR_node.get_attr('beta'),
            IR_node.get_attr('k')
        ))


    def emit_Add(self, IR_node):
        input_layers = ', '.join(('n.' + self.IR_graph.get_parent(IR_node.name, [num]).real_variable_name) for num in range(0, len(IR_node.in_edges)))
        self.add_body(1, "n.{:<15} = L.Eltwise({}, operation=1, ntop=1)".format(
            IR_node.variable_name,
            input_layers,
        ))

    def emit_Flatten(self, IR_node):
        self.add_body(1, "n.{:<15} = L.Flatten(n.{})".format(
            IR_node.variable_name,
            self.parent_variable_name(IR_node),
            ))


    def emit_Squeeze(self, IR_node):
        shape = IR_node.get_attr("_output_shapes")[0]
        shape = shape_to_list(shape)
        if shape:
            dim_str = "'dim': {}".format(shape)
            dim_str = " reshape_param={'shape': { " + dim_str + '} }'
            self.add_body(1, "n.{:<15} = L.Reshape(n.{}, {})".format(
                IR_node.variable_name,
                self.parent_variable_name(IR_node),
                dim_str
                ))
        else:
            IR_node.real_name = self.IR_graph.get_parent(IR_node.name, [0]).real_name


    def emit_Concat(self, IR_node):
        axis_array = (2, 3, 1, 0)
        axis = axis_array.index(IR_node.get_attr('axis'))
        input_layers = ', '.join(('n.' + self.IR_graph.get_node(edge).real_variable_name) for edge in IR_node.in_edges)
        self.add_body(1, "n.{:<15} = L.Concat({}, axis={})".format(
            IR_node.variable_name,
            input_layers,
            axis
        ))

    def emit_Sigmoid(self, IR_node):
        self.add_body(1, "n.{:<15} = L.Sigmoid(n.{}, ntop=1)".format(
            IR_node.variable_name,
            self.parent_variable_name(IR_node)))


    def emit_Relu(self, IR_node):
        in_place = True
        self.add_body(1, "n.{:<15} = L.ReLU(n.{}, in_place={}, ntop=1)".format(
            IR_node.variable_name,
            self.parent_variable_name(IR_node),
            in_place))

    def emit_LeakyRelu(self, IR_node):
        in_place = True
        self.add_body(1, "n.{:<15} = L.ReLU(n.{}, in_place={}, negative_slope={}, ntop=1)".format(
            IR_node.variable_name,
            self.parent_variable_name(IR_node),
            in_place,
            IR_node.IR_layer.attr['alpha'].f))



    def emit_PRelu(self, IR_node):
        in_place = True
        self.add_body(1, "n.{:<15} = L.PReLU(n.{}, in_place={}, ntop=1)".format(
            IR_node.variable_name,
            self.parent_variable_name(IR_node),
            in_place))

    def emit_Tanh(self, IR_node):
        self.add_body(1, "n.{:<15} = L.TanH(n.{}, ntop=1)".format(
            IR_node.variable_name,
            self.parent_variable_name(IR_node)))

    def emit_Softmax(self, IR_node):
        self.add_body(1, "n.{:<15} = L.Softmax(n.{}, ntop=1)".format(
            IR_node.variable_name,
            self.parent_variable_name(IR_node)))


    def emit_Pad(self, IR_node):
        IR_node.real_name = self.IR_graph.get_parent(IR_node.name, [0]).real_name

    def reduction(self, IR_node, op, axes):
        # Convert NHWC (IR) to NCHW (Caffe): [0,1,2,3]->[0,3,1,2]
        if len(axes) == 1:
            assert (axes[0] == 2)
        elif len(axes) == 2:
            assert ((axes[0] == 1) and (axes[1] == 2))

        self.add_body(1, "n.{:<15} = L.Reduction(n.{}, operation={} , axis={} ,ntop=1)".format(
            IR_node.variable_name,
            self.parent_variable_name(IR_node),
            op,
            len(axes)))

        if IR_node.get_attr('keepdims') == True:
            shape = IR_node.get_attr("_output_shapes")[0]
            shape = shape_to_list(shape)
            shape = [1] + [shape[-1]] + shape[1:-1]
            dim_str = "'dim': {}".format(shape)
            dim_str = "{'shape': { " + dim_str + '} }'
            self.add_body(1, "n.{:<15} = L.Reshape(n.{}, reshape_param={}) ".format(
                IR_node.variable_name + "_reshape",
                IR_node.real_variable_name,
                dim_str))
            IR_node.real_name = IR_node.real_name + '_reshape'


    def emit_ReduceMean(self, IR_node):
        self.reduction(IR_node, 4, IR_node.get_attr('axes'))

    def emit_ReduceSum(self, IR_node):
        self.reduction(IR_node, 1, IR_node.get_attr('axes'))

    def emit_Relu6(self, IR_node):
        self.emit_Relu(IR_node)

    def emit_DepthwiseConv(self, IR_node):
        self.emit_Conv(IR_node)

    def emit_Const(self, IR_node):
        pass
    def emit_Shape(self, IR_node):
        pass
    def emit_Reshape(self, IR_node):
        shape = IR_node.get_attr("_output_shapes")[0]
        shape = shape_to_list(shape)
        if shape:
            dim_str = "'dim': {}".format(shape)
            dim_str = " reshape_param={'shape': { " + dim_str + '} }'
            self.add_body(1, "n.{:<15} = L.Reshape(n.{}, {})".format(
                IR_node.variable_name,
                self.parent_variable_name(IR_node),
                dim_str
                ))
        else:
            IR_node.real_name = self.IR_graph.get_parent(IR_node.name, [0]).real_name

    def emit_Slice(self, IR_node):
        pass

    def emit_Pack(self, IR_node):
        pass

    def emit_Abs(self, IR_node):
        self.add_body(1, "n.{:<15} = L.AbsVal(n.{}, ntop=1)".format(
            IR_node.variable_name,
            self.parent_variable_name(IR_node)))

    def emit_Sub(self, IR_node):
        input_layers = ', '.join(('n.' + self.IR_graph.get_node(edge).real_variable_name) for edge in IR_node.in_edges)
        self.add_body(1, "n.{:<15} = L.Eltwise({}, coeff = [1, -1], ntop=1)".format(
            IR_node.variable_name,
            input_layers))

    def emit_Mul(self, IR_node):
        if len(IR_node.in_edges) == 2:
            input_layers = ', '.join(('n.' + self.IR_graph.get_node(edge).real_variable_name) for edge in IR_node.in_edges)
            self.add_body(1, "n.{:<15} = L.Eltwise({}, operation=0, ntop=1)".format(
            IR_node.variable_name,
            input_layers))
        elif len(IR_node.in_edges) == 1:
            self.emit_Scale(IR_node)
        else:
            assert False

    def emit_UpSampling2D(self, IR_node):
        scales = IR_node.get_attr('scales')
        scale = tuple(scales)[0]

        shape = IR_node.get_attr('_output_shapes')[0]
        shape = shape_to_list(shape)

        self.add_body(1, "n.{:<15} = L.Deconvolution(n.{}, convolution_param=dict(kernel_size={}, stride={}, pad={}, num_output={}, group={}, bias_term={}), param=[dict(lr_mult=0)], ntop=1)".format(
            IR_node.variable_name,
            IR_node.in_edges[0],
            2 * scale - scale % 2,
            scale,
            int(math.ceil((scale - 1) / 2)),
            shape[-1],
            shape[-1],
            False))

    # def emit_Square(self, IR_node):
    #     input_layers = ', '.join(('n.' + self.IR_graph.get_node(edge).real_variable_name) for edge in IR_node.in_edges)
    #     self.add_body(1, "n.{:<15} = L.Square({}, ntop=1)".format(
    #         IR_node.variable_name,
    #         input_layers))


    def emit_Elu(self, IR_node):
        in_place = True
        self.add_body(1, "n.{:<15} = L.ELU(n.{}, in_place={}, ntop=1)".format(
            IR_node.variable_name,
            self.parent_variable_name(IR_node),
            in_place))


#----------------------------------------------------------------------------------------------
#  Copyright (c) Microsoft Corporation. All rights reserved.
#  Licensed under the MIT License. See License.txt in the project root for license information.
#----------------------------------------------------------------------------------------------

from __future__ import absolute_import

import os
import tensorflow as tf

from tensorflow.contrib.slim.nets import vgg
from tensorflow.contrib.slim.nets import inception
from tensorflow.contrib.slim.nets import resnet_v1
from tensorflow.contrib.slim.nets import resnet_v2
from mmdnn.conversion.examples.tensorflow.models import inception_resnet_v2
from mmdnn.conversion.examples.tensorflow.models import mobilenet_v1
from mmdnn.conversion.examples.tensorflow.models import nasnet
from mmdnn.conversion.examples.tensorflow.models.mobilenet import mobilenet_v2
from mmdnn.conversion.examples.tensorflow.models import inception_resnet_v1
from mmdnn.conversion.examples.tensorflow.models import test_rnn
slim = tf.contrib.slim
from mmdnn.conversion.examples.imagenet_test import TestKit
from mmdnn.conversion.examples.extractor import base_extractor
from mmdnn.conversion.common.utils import download_file


class tensorflow_extractor(base_extractor):

    architecture_map = {
        'vgg16' : {
            'url'         : 'http://download.tensorflow.org/models/vgg_16_2016_08_28.tar.gz',
            'filename'    : 'vgg_16.ckpt',
            'builder'     : lambda : vgg.vgg_16,
            'arg_scope'   : vgg.vgg_arg_scope,
            'input'       : lambda : tf.placeholder(name='input', dtype=tf.float32, shape=[None, 224, 224, 3]),
            'num_classes' : 1000,
        },
        'vgg19' : {
            'url'         : 'http://download.tensorflow.org/models/vgg_19_2016_08_28.tar.gz',
            'filename'    : 'vgg_19.ckpt',
            'builder'     : lambda : vgg.vgg_19,
            'arg_scope'   : vgg.vgg_arg_scope,
            'input'       : lambda : tf.placeholder(name='input', dtype=tf.float32, shape=[None, 224, 224, 3]),
            'num_classes' : 1000,
        },
        'inception_v1' : {
            'url'         : 'http://download.tensorflow.org/models/inception_v1_2016_08_28.tar.gz',
            'filename'    : 'inception_v1.ckpt',
            'builder'     : lambda : inception.inception_v1,
            'arg_scope'   : inception.inception_v3_arg_scope,
            'input'       : lambda : tf.placeholder(name='input', dtype=tf.float32, shape=[None, 224, 224, 3]),
            'num_classes' : 1001,
        },
        'inception_v1_frozen' : {
            'url'         : 'https://storage.googleapis.com/download.tensorflow.org/models/inception_v1_2016_08_28_frozen.pb.tar.gz',
            'filename'    : 'inception_v1_2016_08_28_frozen.pb',
            'tensor_out'  : ['InceptionV1/Logits/Predictions/Reshape_1:0'],
            'tensor_in'   : ['input:0'],
            'input_shape' : [[224, 224, 3]],  # input_shape of the elem in tensor_in
            'feed_dict'   :lambda img: {'input:0':img},
            'num_classes' : 1001,
        },
        'inception_v3' : {
            'url'         : 'http://download.tensorflow.org/models/inception_v3_2016_08_28.tar.gz',
            'filename'    : 'inception_v3.ckpt',
            'builder'     : lambda : inception.inception_v3,
            'arg_scope'   : inception.inception_v3_arg_scope,
            'input'       : lambda : tf.placeholder(name='input', dtype=tf.float32, shape=[None, 299, 299, 3]),
            'num_classes' : 1001,
        },
        'inception_v3_frozen' : {
            'url'         : 'https://storage.googleapis.com/download.tensorflow.org/models/inception_v3_2016_08_28_frozen.pb.tar.gz',
            'filename'    : 'inception_v3_2016_08_28_frozen.pb',
            'tensor_out'  : ['InceptionV3/Predictions/Softmax:0'],
            'tensor_in'   : ['input:0'],
            'input_shape' : [[299, 299, 3]], # input_shape of the elem in tensor_in
            'feed_dict'   :lambda img: {'input:0':img},
            'num_classes' : 1001,
        },
        'resnet_v1_50' : {
            'url'         : 'http://download.tensorflow.org/models/resnet_v1_50_2016_08_28.tar.gz',
            'filename'    : 'resnet_v1_50.ckpt',
            'builder'     : lambda : resnet_v1.resnet_v1_50,
            'arg_scope'   : resnet_v2.resnet_arg_scope,
            'input'       : lambda : tf.placeholder(name='input', dtype=tf.float32, shape=[None, 224, 224, 3]),
            'num_classes' : 1000,
        },
        'resnet_v1_152' : {
            'url'         : 'http://download.tensorflow.org/models/resnet_v1_152_2016_08_28.tar.gz',
            'filename'    : 'resnet_v1_152.ckpt',
            'builder'     : lambda : resnet_v1.resnet_v1_152,
            'arg_scope'   : resnet_v2.resnet_arg_scope,
            'input'       : lambda : tf.placeholder(name='input', dtype=tf.float32, shape=[None, 224, 224, 3]),
            'num_classes' : 1000,
        },
        'resnet_v2_50' : {
            'url'         : 'http://download.tensorflow.org/models/resnet_v2_50_2017_04_14.tar.gz',
            'filename'    : 'resnet_v2_50.ckpt',
            'builder'     : lambda : resnet_v2.resnet_v2_50,
            'arg_scope'   : resnet_v2.resnet_arg_scope,
            'input'       : lambda : tf.placeholder(name='input', dtype=tf.float32, shape=[None, 299, 299, 3]),
            'num_classes' : 1001,
        },
        'resnet_v2_101' : {
            'url'         : 'http://download.tensorflow.org/models/resnet_v2_101_2017_04_14.tar.gz',
            'filename'    : 'resnet_v2_101.ckpt',
            'builder'     : lambda : resnet_v2.resnet_v2_101,
            'arg_scope'   : resnet_v2.resnet_arg_scope,
            'input'       : lambda : tf.placeholder(name='input', dtype=tf.float32, shape=[None, 299, 299, 3]),
            'num_classes' : 1001,
        },
        'resnet_v2_152' : {
            'url'         : 'http://download.tensorflow.org/models/resnet_v2_152_2017_04_14.tar.gz',
            'filename'    : 'resnet_v2_152.ckpt',
            'builder'     : lambda : resnet_v2.resnet_v2_152,
            'arg_scope'   : resnet_v2.resnet_arg_scope,
            'input'       : lambda : tf.placeholder(name='input', dtype=tf.float32, shape=[None, 299, 299, 3]),
            'num_classes' : 1001,
        },
        'resnet_v2_200' : {
            'url'         : 'http://download.tensorflow.org/models/resnet_v2_200_2017_04_14.tar.gz',
            'filename'    : 'resnet_v2_200.ckpt',
            'builder'     : lambda : resnet_v2.resnet_v2_200,
            'arg_scope'   : resnet_v2.resnet_arg_scope,
            'input'       : lambda : tf.placeholder(name='input', dtype=tf.float32, shape=[None, 299, 299, 3]),
            'num_classes' : 1001,
        },
        'mobilenet_v1_1.0' : {
            'url'         : 'http://download.tensorflow.org/models/mobilenet_v1_1.0_224_2017_06_14.tar.gz',
            'filename'    : 'mobilenet_v1_1.0_224.ckpt',
            'builder'     : lambda : mobilenet_v1.mobilenet_v1,
            'arg_scope'   : mobilenet_v1.mobilenet_v1_arg_scope,
            'input'       : lambda : tf.placeholder(name='input', dtype=tf.float32, shape=[None, 224, 224, 3]),
            'num_classes' : 1001,
        },
        'mobilenet_v1_1.0_frozen' : {
            'url'         : 'https://storage.googleapis.com/download.tensorflow.org/models/mobilenet_v1_1.0_224_frozen.tgz',
            'filename'    : 'mobilenet_v1_1.0_224/frozen_graph.pb',
            'tensor_out'  : ['MobilenetV1/Predictions/Softmax:0'],
            'tensor_in'   : ['input:0'],
            'input_shape' : [[224, 224, 3]], # input_shape of the elem in tensor_in
            'feed_dict'   :lambda img: {'input:0':img},
            'num_classes' : 1001,
        },
        'mobilenet_v2_1.0_224':{
            'url'         : 'https://storage.googleapis.com/mobilenet_v2/checkpoints/mobilenet_v2_1.0_224.tgz',
            'filename'    : 'mobilenet_v2_1.0_224.ckpt',
            'builder'     : lambda : mobilenet_v2.mobilenet,
            'arg_scope'   : mobilenet_v2.training_scope,
            'input'       : lambda : tf.placeholder(name='input', dtype=tf.float32, shape=[None, 224, 224, 3]),
            'num_classes' : 1001,
        },
        'inception_resnet_v2' : {
            'url'         : 'http://download.tensorflow.org/models/inception_resnet_v2_2016_08_30.tar.gz',
            'filename'    : 'inception_resnet_v2_2016_08_30.ckpt',
            'builder'     : lambda : inception_resnet_v2.inception_resnet_v2,
            'arg_scope'   : inception_resnet_v2.inception_resnet_v2_arg_scope,
            'input'       : lambda : tf.placeholder(name='input', dtype=tf.float32, shape=[None, 299, 299, 3]),
            'num_classes' : 1001,
        },
        'nasnet-a_large' : {
            'url'         : 'https://storage.googleapis.com/download.tensorflow.org/models/nasnet-a_large_04_10_2017.tar.gz',
            'filename'    : 'model.ckpt',
            'builder'     : lambda : nasnet.build_nasnet_large,
            'arg_scope'   : nasnet.nasnet_large_arg_scope,
            'input'       : lambda : tf.placeholder(name='input', dtype=tf.float32, shape=[None, 331, 331, 3]),
            'num_classes' : 1001,
        },
        'facenet' : {
            'url'         : 'http://mmdnn.eastasia.cloudapp.azure.com:89/models/tensorflow/facenet/20180408-102900.zip',
            'filename'    : '20180408-102900/model-20180408-102900.ckpt-90',
            'builder'     : lambda : inception_resnet_v1.inception_resnet_v1,
            'arg_scope'   : inception_resnet_v1.inception_resnet_v1_arg_scope,
            'input'       : lambda : tf.placeholder(name='input', dtype=tf.float32, shape=[None, 160, 160, 3]),
            'feed_dict'   : lambda img: {'input:0':img,'phase_train:0':False},
            'num_classes' : 0,
        },
        'facenet_frozen' : {
            'url'         : 'http://mmdnn.eastasia.cloudapp.azure.com:89/models/tensorflow/facenet/20180408-102900.zip',
            'filename'    : '20180408-102900/20180408-102900.pb',
            'tensor_out'  : ['InceptionResnetV1/Logits/AvgPool_1a_8x8/AvgPool:0'],
            'tensor_in'   : ['input:0','phase_train:0'],
            'input_shape' : [[160, 160, 3],1], # input_shape of the elem in tensor_in
            'feed_dict'   : lambda img: {'input:0':img,'phase_train:0':False},
            'num_classes' : 0,
        },
        'rnn_lstm_gru_stacked': {
            'url'         :'http://mmdnn.eastasia.cloudapp.azure.com:89/models/tensorflow/tf_rnn/tf_rnn.zip',  # Note this is just a model used for test, not a standard rnn model.
            'filename'    :'tf_rnn/tf_lstm_gru_stacked.ckpt',
            'builder'     :lambda: test_rnn.create_symbol,
            'arg_scope'   :test_rnn.dummy_arg_scope,
            'input'       :lambda: tf.placeholder(name='input', dtype=tf.int32, shape=[None, 150]),
            'feed_dict'   :lambda x:{'input:0': x},
            'num_classes' : 0
        }
    }


    @classmethod
    def handle_checkpoint(cls, architecture, path):
        with slim.arg_scope(cls.architecture_map[architecture]['arg_scope']()):
            data_input = cls.architecture_map[architecture]['input']()
            logits, endpoints = cls.architecture_map[architecture]['builder']()(
                data_input,
                num_classes=cls.architecture_map[architecture]['num_classes'],
                is_training=False)

            if logits.op.type == 'Squeeze':
                labels = tf.identity(logits, name='MMdnn_Output')
            else:
                labels = tf.squeeze(logits, name='MMdnn_Output')
        

        init = tf.global_variables_initializer()
        with tf.Session() as sess:
            sess.run(init)
            saver = tf.train.Saver()
            saver.restore(sess, path + cls.architecture_map[architecture]['filename'])
            save_path = saver.save(sess, path + "imagenet_{}.ckpt".format(architecture))
            print("Model saved in file: %s" % save_path)

        import tensorflow.contrib.keras as keras
        keras.backend.clear_session()


    @classmethod
    def handle_frozen_graph(cls, architecture, path):
        return
        # raise NotImplementedError()

    @classmethod
    def get_frozen_para(cls, architecture):
        frozenname = architecture + '_frozen'
        tensor_in =  list(map(lambda x:x.split(':')[0], cls.architecture_map[frozenname]['tensor_in']))
        tensor_out = list(map(lambda x:x.split(':')[0], cls.architecture_map[frozenname]['tensor_out']))
        return cls.architecture_map[frozenname]['filename'], cls.architecture_map[frozenname]['input_shape'], tensor_in, tensor_out


    @classmethod
    def download(cls, architecture, path="./"):
        if cls.sanity_check(architecture):
            architecture_file = download_file(cls.architecture_map[architecture]['url'], directory=path, auto_unzip=True)
            if not architecture_file:
                return None

            tf.reset_default_graph()

            if 'ckpt' in cls.architecture_map[architecture]['filename']:
                cls.handle_checkpoint(architecture, path)

            elif cls.architecture_map[architecture]['filename'].endswith('pb'):
                cls.handle_frozen_graph(architecture, path)
            
            else:
                raise ValueError("Unknown file name [{}].".format(cls.architecture_map[architecture]['filename']))

            return architecture_file

        else:
            return None


    @classmethod
    def inference(cls, architecture, files, path, test_input_path, is_frozen=False):
        if is_frozen:
            architecture_ = architecture + "_frozen"
        else:
            architecture_ = architecture

        if cls.download(architecture_, path):
            import numpy as np
            if 'rnn' not in architecture_:
                func = TestKit.preprocess_func['tensorflow'][architecture]
                img = func(test_input_path)
                img = np.expand_dims(img, axis=0)
                input_data = img
            else:
                input_data = np.load(test_input_path)

            if is_frozen:
                tf_model_path = cls.architecture_map[architecture_]['filename']
                with open(path + tf_model_path, 'rb') as f:
                    serialized = f.read()
                tf.reset_default_graph()
                original_gdef = tf.GraphDef()
                original_gdef.ParseFromString(serialized)
                tf_output_name =  cls.architecture_map[architecture_]['tensor_out']
                tf_input_name =  cls.architecture_map[architecture_]['tensor_in']
                feed_dict = cls.architecture_map[architecture_]['feed_dict']

                with tf.Graph().as_default() as g:
                    tf.import_graph_def(original_gdef, name='')
                with tf.Session(graph = g) as sess:
                    tf_out = sess.run(tf_output_name[0], feed_dict=feed_dict(input_data)) # temporarily think the num of out nodes is one
                predict = np.squeeze(tf_out)
                return predict

            else:
                with slim.arg_scope(cls.architecture_map[architecture]['arg_scope']()):
                    data_input = cls.architecture_map[architecture]['input']()
                    logits, endpoints = cls.architecture_map[architecture]['builder']()(
                        data_input,
                        num_classes=cls.architecture_map[architecture]['num_classes'],
                        is_training=False)
                    labels = tf.squeeze(logits)

                init = tf.global_variables_initializer()
                with tf.Session() as sess:
                    sess.run(init)
                    saver = tf.train.Saver()
                    saver.restore(sess, path + cls.architecture_map[architecture]['filename'])
                    predict = sess.run(logits, feed_dict = {data_input : input_data})

                import tensorflow.contrib.keras as keras
                keras.backend.clear_session()

                predict = np.squeeze(predict)
                return predict

        else:
            return None




#----------------------------------------------------------------------------------------------
#  Copyright (c) Microsoft Corporation. All rights reserved.
#  Licensed under the MIT License. See License.txt in the project root for license information.
#----------------------------------------------------------------------------------------------


import os
from six import string_types as _string_types
import numpy as np
import math

from coremltools.models.neural_network import NeuralNetworkBuilder as _NeuralNetworkBuilder
from coremltools.models import datatypes
from coremltools.models import MLModel as _MLModel
from coremltools.models.utils import save_spec as _save_spec
from coremltools.models._infer_shapes_nn_mlmodel import infer_shapes
from coremltools.proto import Model_pb2 ,NeuralNetwork_pb2


from mmdnn.conversion.coreml.coreml_graph import CoremlGraph
import mmdnn.conversion.common.IR.graph_pb2 as graph_pb2
from mmdnn.conversion.common.IR.graph_pb2 import NodeDef, GraphDef, DataType
from mmdnn.conversion.common.DataStructure.parser import Parser
from mmdnn.conversion.common.utils import *


class CoremlParser(Parser):

    activation_map = {
        "ReLU"          : "Relu",
        "leakyReLU"     : "LeakyRelu",
        "linear"        : "linear",
        "thresholdedReLU" : "ThresholdedRelu",
        "PReLU"         : "PRelu",
        "tanh"          : "Tanh",
        "scaledTanh"    : "ScaledTanh",
        'sigmoid'       : "Sigmoid",
        "sigmoidHard"   : "SigmoidHard",
        "ELU"           : "Elu",
        'softplus'      : 'Softplus',
        'softsign'      : 'Softsign',
        'parametricSoftplus'    : "ParametricSoftplus"
    }




    def __init__(self, model):
        super(CoremlParser, self).__init__()

        # load model file into Coreml Graph
        if isinstance(model, _string_types):
            # model.encode() convert to str --- python2 may crash due to type 'unicode'
            model = _MLModel(model)
            model = model.get_spec()
            self.weight_loaded = True
        else:
            assert False

        # Build Network Graph

        model_type = model.WhichOneof('Type')
        if model_type == 'neuralNetworkClassifier':
            CoremlParser.shape_dict = infer_shapes(model.neuralNetworkClassifier, model.description.input)
        elif model_type == 'neuralNetwork':
            CoremlParser.shape_dict = infer_shapes(model.neuralNetwork, model.description.input)
        elif model_type == 'neuralNetworkRegressor':
            CoremlParser.shape_dict = infer_shapes(model.neuralNetworkRegressor, model.description.input)
        else:
            assert False

        # self.data_format ? TODO
        self.data_format = 'channels_first'
        self.coreml_graph = CoremlGraph(model)
        self.coreml_graph.build()
        self.lambda_layer_count = 0


    def _load_model(self, model_network_path):
        """Load a Coreml model from disk

        Parameters
        ----------

        model_network_path: str
            Path where the model network path is (mlmodel file)

        Returns
        -------
        model: A coreml model
        """

        from coremltools.models import MLModel

        if os.path.isfile(model_network_path):
            # load the model network
            loaded_model_ml = MLModel(model_network_path)
            # convert to Model_pb2.Model
            loaded_model_pb = loaded_model_ml.get_spec()
            self.weight_loaded = True
            print("Network file [{}] is loaded successfully.".format(model_network_path))
        else:
            print("Warning: Weights File [{}] is not found.".format(model_network_path))

        return loaded_model_pb


    @property
    def src_graph(self):
        return self.coreml_graph



    def gen_IR(self):
        for i, layer in enumerate(self.coreml_graph.topological_sort):

            current_node = self.coreml_graph.get_node(layer)
            current_node_layer = current_node.layer

            # determine the type of the current_node
            node_type = current_node_layer.name

            if isinstance(current_node_layer, Model_pb2.FeatureDescription):
                self.rename_InputLayer(current_node)
            elif isinstance(current_node_layer, NeuralNetwork_pb2.NeuralNetworkLayer):
                if current_node_layer.HasField("convolution"):
                    self.rename_CONV2D(current_node)
                elif current_node_layer.HasField('batchnorm'):
                    self.rename_BatchNormalization(current_node)
                elif current_node_layer.HasField("scale"):
                    self.rename_scale(current_node)
                elif current_node_layer.HasField("pooling"):
                    self.rename_Pooling(current_node)
                elif current_node_layer.HasField("activation"):
                    self.rename_Activation(current_node)
                elif current_node_layer.HasField("softmax"):
                    self.rename_Softmax(current_node)
                elif current_node_layer.HasField("padding"):
                    self.rename_Padding(current_node)
                elif current_node_layer.HasField("add"):
                    self.rename_Add(current_node)
                elif current_node_layer.HasField("flatten"):
                    self.rename_Flatten(current_node)
                elif current_node_layer.HasField("innerProduct"):
                    self.rename_innerProduct(current_node)
                elif current_node_layer.HasField("concat"):
                    self.rename_Concatenate(current_node)
                else:
                    print("CoremlParser has not supported operator [{}]".format(node_type))
                    self.rename_UNKNOWN(current_node)
            else:
                assert False



    # staticmethods
    @staticmethod
    def _set_output_shape(source_node, IR_node):

        shape = graph_pb2.TensorShape()
        source_node_layer = source_node.layer

        layer_name = source_node_layer.output[0]

        shape_coreml = CoremlParser.shape_dict[layer_name]
        # (seq, batch, C, H, W)  & NHWC

        new_dim = shape.dim.add()
        if shape_coreml[1] == 1:
            new_dim.size = -1
        else:
            new_dim.size = shape_coreml[1]
        for index in [3, 4, 2]:
            new_dim = shape.dim.add()
            dim = shape_coreml[index]
            new_dim.size = dim if dim else -1

        IR_node.attr["_output_shapes"].list.shape.extend([shape])


    @staticmethod
    def _copy_and_reop(source_node, IR_node, new_op = None):
        source_node_layer = source_node.layer
        IR_node.name = source_node_layer.name

        if new_op:
            IR_node.op = new_op
        elif source_node_layer.HasField("convolution"):
            IR_node.op = "convolution"
        elif source_node_layer.HasField('batchnorm'):
            IR_node.op = "batchnorm"
        elif source_node_layer.HasField("scale"):
            IR_node.op = "scale"
        elif source_node_layer.HasField("pooling"):
            IR_node.op = "pooling"
        elif source_node_layer.HasField("activation"):
            IR_node.op = "activation"
        elif source_node_layer.HasField("softmax"):
            IR_node.op = "softmax"
        elif source_node_layer.HasField("padding"):
            IR_node.op = "padding"
        elif source_node_layer.HasField("add"):
            IR_node.op = "add"
        elif source_node_layer.HasField("flatten"):
            IR_node.op = "flatten"
        elif source_node_layer.HasField("innerProduct"):
            IR_node.op = "innerProduct"
        elif source_node_layer.HasField("concat"):
            IR_node.op = "concatenate"
        else:
            assert False

        #  TODO dtype_map
        if hasattr(source_node.layer, "dtype"):
            IR_node.attr["dtype"].type = CoremlParser.dtype_map[source_node.layer.dtype]

        CoremlParser._set_output_shape(source_node, IR_node)

    @staticmethod
    def _copy_shape(source_node, target_node):
        if hasattr(source_node, "output_shape"):
            for dim in source_node.output_shape:
                new_dim = target_node.attr['shape'].shape.dim.add()
                new_dim.size = -1 if dim == None else new_dim
        else:
            target_node.attr['shape'].shape.unknown_rank = True

    @staticmethod
    def _convert_dataformat(source_node, target_node):
        if source_node.coreml_layer.data_format == "channels_last":
            target_node.attr['data_format'].s = "NHWC"
        elif source_node.coreml_layer.data_format == 'channels_first':
            target_node.attr['data_format'].s = "NCHW"
        else:
            print("Warning: [%s] don't have data format info" % (source_node.coreml_layer.name))


###### convert methods

    # convolution
    def __convert_convolution(self, source_node, dim):

        IR_node = self.IR_graph.node.add()
        # input edge
        self.convert_inedge(source_node, IR_node)
        source_node_layer = source_node.layer
        source_node_conv = source_node_layer.convolution
        layer_name = source_node_layer.name.split('/')[-1]

        # important!
        if source_node_conv.HasField('weights'):
            # reshape the weight!
            [h , w , k , o] = list(source_node_conv.kernelSize) + [source_node_conv.kernelChannels , source_node_conv.outputChannels]
            # [2, 3, 0, 1]
            weights = np.array(source_node_conv.weights.floatValue, dtype=np.float32).reshape([o, k, h, w]).transpose([2, 3, 1, 0])



        kwargs = dict()
        kwargs['kernel_shape'] = list(source_node_conv.kernelSize) + [source_node_conv.kernelChannels, source_node_conv.outputChannels]

        # pads
        CoremlParser._convert_padding(source_node, IR_node)
        # use_bias
        kwargs['use_bias'] = source_node_conv.hasBias
        # isDeconvolution
        kwargs['isDeconvolution'] = source_node_conv.isDeconvolution
        # name, op
        if layer_name == 'sep':
            CoremlParser._copy_and_reop(source_node, IR_node, "Conv")
        elif layer_name == 'dw':
            CoremlParser._copy_and_reop(source_node, IR_node, "DepthwiseConv")
            weights = weights.transpose((0,1,3,2))
            kwargs['kernel_shape'] = list(source_node_conv.kernelSize) + [source_node_conv.outputChannels, source_node_conv.kernelChannels]


        else:
            if kwargs['isDeconvolution']:
                CoremlParser._copy_and_reop(source_node, IR_node, "ConvTranspose")
            else:
                CoremlParser._copy_and_reop(source_node, IR_node, "Conv")

        self.set_weight(source_node.name, 'weights',  weights)
        if source_node_layer.convolution.HasField('bias'):
            self.set_weight(source_node.name, 'bias', np.array(source_node_conv.bias.floatValue, dtype=np.float32))



        # kwargs['kernel_shape'] = weights.shape

        kwargs['group'] = source_node_conv.nGroups

        # strides
        # [1, sd, sh, sw, 1]
        kwargs['strides'] = [1] + list(source_node_conv.stride) + [1]

        dilation = list(source_node_conv.dilationFactor)
        if dilation == []:
            dilation = [1,1]
        kwargs['dilations'] = [1] + dilation + [1]


        assign_IRnode_values(IR_node, kwargs)


        # activation
        # TODO
        self._defuse_activation(source_node)



    @staticmethod
    def _convert_padding(source_node, IR_node):
        source_node_layer = source_node.layer

        if source_node_layer.HasField('convolution'):
            # padding in conv

            source_node_conv = source_node_layer.convolution


            if source_node_conv.HasField('valid'):
                # pad in IR is [x1_b, x2_b, ..., x1_e, x2_e, ...]

                dim = []
                for i in source_node_conv.valid.paddingAmounts.borderAmounts:
                    dim.extend([i.startEdgeSize, i.endEdgeSize])


                if dim == []:
                    assign_IRnode_values(IR_node, { 'auto_pad': 'VALID'})
                    pad_dim = [0] * 8
                    pad_dim = convert_tf_pad_to_onnx(pad_dim)

                    assign_IRnode_values(IR_node, { 'pads':  pad_dim})
                else:

                    # padding
                    pad_dim = [0, 0]

                    pad_dim.extend(dim)

                    pad_dim += [0, 0]

                    pad_dim = convert_tf_pad_to_onnx(pad_dim)


                    assign_IRnode_values(IR_node, { 'pads':  pad_dim})

            elif source_node_conv.HasField('same'):

                # compute padding for 'same'
                assign_IRnode_values(IR_node, {'auto_pad': "SAME"})


                kernel = list(source_node_conv.kernelSize)
                dilation = list(source_node_conv.dilationFactor)
                if dilation == []:
                    dilation = [1,1]
                stride = list(source_node_conv.stride)
                if stride == []:
                    stride = [1,1]

                kernel[0] = dilation[0] * ( kernel[0] -1 ) + 1
                kernel[1] = dilation[1] * ( kernel[1] -1 ) + 1


                if stride == [1,1]:

                    # https://discuss.mxnet.io/t/pooling-and-convolution-with-same-mode/528/3

                    p0 =  ( kernel[0] -1 ) // 2
                    p1 =  ( kernel[1] -1 ) // 2

                    if kernel[0] % 2 == 0:
                        p00 = p0
                        p01 = p0 + 1
                    else:
                        p00 = p0
                        p01 = p0

                    if kernel[1] % 2 == 0:
                        p10 = p1
                        p11 = p1 + 1
                    else:
                        p10 = p1
                        p11 = p1

                    pad_dim = [0, 0, p00, p01, p10, p11, 0, 0]


                    pad_dim = convert_tf_pad_to_onnx(pad_dim)

                    assign_IRnode_values(IR_node, { 'pads':  pad_dim})
                else:
                    # https://www.jianshu.com/p/05c4f1621c7e
                    pad_dim = [0, 0, 0, 0, 0, 0, 0, 0]

                    pad_dim = convert_tf_pad_to_onnx(pad_dim)

                    assign_IRnode_values(IR_node, { 'pads':  pad_dim})

            else:
                assert False

        elif source_node_layer.HasField('pooling'):
            # padding in pooling
            source_node_pool = source_node_layer.pooling
            if  source_node_pool.HasField('valid'):

                dim = []
                for i in source_node_pool.valid.paddingAmounts.borderAmounts:
                    dim.extend([i.startEdgeSize, i.endEdgeSize])


                if dim == []:
                    assign_IRnode_values(IR_node, { 'auto_pad': 'VALID'})
                    pad_dim = [0] * 8
                    pad_dim = convert_tf_pad_to_onnx(pad_dim)

                    assign_IRnode_values(IR_node, { 'pads':  pad_dim})
                else:
                    # padding
                    pad_dim = [0, 0]

                    pad_dim.extend(dim)

                    pad_dim += [0, 0]
                    pad_dim = convert_tf_pad_to_onnx(pad_dim)
                    assign_IRnode_values(IR_node, { 'pads':  pad_dim})


            elif source_node_pool.HasField('same'):

                assign_IRnode_values(IR_node, { 'auto_pad': 'SAME'})

                kernel = list(source_node_pool.kernelSize)
                stride = list(source_node_pool.stride)
                if stride == []:
                    stride = [1,1]


                if stride == [1,1]:
                    # https://discuss.mxnet.io/t/pooling-and-convolution-with-same-mode/528/3
                    p0 =  ( kernel[0] -1 ) // 2
                    p1 =  ( kernel[1] -1 ) // 2



                    if kernel[0] % 2 == 0:
                        p00 = p0
                        p01 = p0 + 1
                    else:
                        p00 = p0
                        p01 = p0

                    if kernel[1] % 2 == 0:
                        p10 = p1
                        p11 = p1 + 1
                    else:
                        p10 = p1
                        p11 = p1

                    pad_dim = [0, 0, p00, p01, p10, p11, 0, 0]



                    pad_dim = convert_tf_pad_to_onnx(pad_dim)

                    assign_IRnode_values(IR_node, { 'pads':  pad_dim})
                else:
                    # TODO
                    pad_dim = [0, 0, 0, 0, 0, 0, 0, 0]

                    pad_dim = convert_tf_pad_to_onnx(pad_dim)

                    assign_IRnode_values(IR_node, { 'pads':  pad_dim})

            elif source_node_pool.HasField('includeLastPixel'):

                # symmetric padding
                h, w = source_node_pool.includeLastPixel.paddingAmounts
                assign_IRnode_values(IR_node, { 'pads':  [ 0,h, h,0,0, w, w,0]})
            else:
                assert False

        else:
            assert False




    def _convert_merge(self, source_node, new_name = None):

        IR_node = self.IR_graph.node.add()

        # name, op
        CoremlParser._copy_and_reop(source_node, IR_node, new_name)

        # input edge
        self.convert_inedge(source_node, IR_node)

        # For concat axis
        # NO axis in coreml, so set the last axis
        IR_node.attr['axis'].i = len(CoremlParser.shape_dict[source_node.layer.output[0]])-1 -1
        # The first -1 means in coreml there is one-more axis,
        # The second -1 means the last axis

        return IR_node


    def _convert_padding_api(self, source_node, IR_node):
        # name, op
        CoremlParser._copy_and_reop(source_node, IR_node, "Pad")

        # input edge
        self.convert_inedge(source_node, IR_node)

        kwargs = dict()

        source_node_layer = source_node.layer
        source_node_pad = source_node_layer.padding

        if source_node_pad.HasField('constant'):
            kwargs['mode'] = 'CONSTANT'
        elif source_node_pad.HasField('reflection'):
            kwargs['mode'] = 'REFLECT'
        elif source_node_pad.HasField('replication'):
            kwargs['mode'] = 'SYMMETRIC'
        else:
            assert False


        dim = []
        for i in source_node_pad.paddingAmounts.borderAmounts:
            dim.extend([i.startEdgeSize, i.endEdgeSize])


        if dim == []:
            dim = [0,0,0,0]

        # padding
        kwargs['pads'] = [0, 0]

        kwargs['pads'].extend(dim)

        kwargs['pads'] += [0, 0]
        kwargs['pads'] = convert_tf_pad_to_onnx(kwargs['pads'])

        assign_IRnode_values(IR_node, kwargs)

    def _defuse_activation(self, source_node):
        # Future Module TODO
        pass
        return


##### rename methods


    def rename_UNKNOWN(self, source_node):
        print(source_node.layer.get_config())
        IR_node = self.IR_graph.node.add()
        CoremlParser._copy_and_reop(source, IR_node)
        self.convert_inedge(source_node, IR_node)



    def rename_Activation(self, coreml_node):
        IR_node = self.IR_graph.node.add()


        coreml_node_layer = coreml_node.layer
        coreml_node_activation = coreml_node_layer.activation

        # name, op
        for activation_name in self.activation_map.keys():
            if coreml_node_activation.HasField(activation_name):
                CoremlParser._copy_and_reop(coreml_node, IR_node, self.activation_map[activation_name])


        # activation type
        activation_type = coreml_node_activation.WhichOneof("NonlinearityType")


        if activation_type == 'leakyReLU':
            assign_IRnode_values(IR_node, {'alpha' : coreml_node_activation.leakyReLU.alpha})
        elif activation_type == 'PReLU':
            assign_IRnode_values(IR_node, {'gamma' : coreml_node_activation.PReLU.alpha})
        elif activation_type == 'ELU':
            assign_IRnode_values(IR_node, {'alpha' : coreml_node_activation.ELU.alpha})
        elif activation_type == 'thresholdedRelu':
            assign_IRnode_values(IR_node, {'alpha' : coreml_node_activation.thresholdedReLU.alpha})
        elif activation_type == 'scaledTanh':
            assign_IRnode_values(IR_node, {'alpha' : coreml_node_activation.scaledTanh.alpha})
            assign_IRnode_values(IR_node, {'beta' : coreml_node_activation.scaledTanh.beta})
        elif activation_type == 'linear':
            assign_IRnode_values(IR_node, {'alpha' : coreml_node_activation.linear.alpha})
            assign_IRnode_values(IR_node, {'beta' : coreml_node_activation.linear.beta})
        elif activation_type == 'sigmoidHard':
            assign_IRnode_values(IR_node, {'alpha' : coreml_node_activation.sigmoidHard.alpha})
            assign_IRnode_values(IR_node, {'beta' : coreml_node_activation.sigmoidHard.beta})
        elif activation_type == 'parametricSoftplus':
            assign_IRnode_values(IR_node, {'alpha' : coreml_node_activation.parametricSoftplus.alpha})
            assign_IRnode_values(IR_node, {'beta' : coreml_node_activation.parametricSoftplus.beta})
        # else:
            # assert False

        # input edge
        self.convert_inedge(coreml_node, IR_node)

    # Merge layers
    def rename_Add(self, source_node):
        self._convert_merge(source_node, 'Add')

    def rename_CONV2D(self, source_node):
        self.__convert_convolution(source_node, 2)


    def rename_InputLayer(self, source_node):
        # only for training
        IR_node = self.IR_graph.node.add()

        # name, op
        IR_node.name = source_node.name
        IR_node.op = "DataInput"
        graph_shape = graph_pb2.TensorShape()
        coreml_node_layer = source_node.layer

        new_dim = graph_shape.dim.add()
        new_dim.size = -1
        new_dim = graph_shape.dim.add()
        new_dim.size = coreml_node_layer.type.imageType.width
        new_dim = graph_shape.dim.add()
        new_dim.size = coreml_node_layer.type.imageType.height
        new_dim = graph_shape.dim.add()

        if coreml_node_layer.type.imageType.colorSpace == 10:
            new_dim.size = 2
        elif coreml_node_layer.type.imageType.colorSpace == 20:
            new_dim.size = 3
        elif coreml_node_layer.type.imageType.colorSpace == 30:
            new_dim.size = 3
        else:
            assert False
        IR_node.attr["_output_shapes"].list.shape.extend([graph_shape])



        # input edge
        self.convert_inedge(source_node, IR_node)




        # shape
        # NHWC channel last
        # in fact, here is NWHC
        new_dim = IR_node.attr['shape'].shape.dim.add()
        new_dim.size = -1
        new_dim = IR_node.attr['shape'].shape.dim.add()
        new_dim.size = coreml_node_layer.type.imageType.width
        new_dim = IR_node.attr['shape'].shape.dim.add()
        new_dim.size = coreml_node_layer.type.imageType.height
        new_dim = IR_node.attr['shape'].shape.dim.add()

        if coreml_node_layer.type.imageType.colorSpace == 10:
            new_dim.size = 2
        elif coreml_node_layer.type.imageType.colorSpace == 20:
            new_dim.size = 3
        elif coreml_node_layer.type.imageType.colorSpace == 30:
            new_dim.size = 3
        else:
            assert False


    def rename_BatchNormalization(self, coreml_node):

        IR_node = self.IR_graph.node.add()

        coreml_node_layer = coreml_node.layer
        coreml_node_bn = coreml_node_layer.batchnorm


        # name, op
        CoremlParser._copy_and_reop(coreml_node, IR_node, "BatchNorm")

        # input edge
        self.convert_inedge(coreml_node, IR_node)


        # axis TODO
        # channels_first, then axis = 1
        IR_node.attr['axis'].i = -1

        # scale
        IR_node.attr['scale'].b = coreml_node_bn.HasField("gamma")

        # bias
        IR_node.attr['bias'].b = coreml_node_bn.HasField("beta")

        # epsilon
        IR_node.attr['epsilon'].f = coreml_node_bn.epsilon

        if IR_node.attr['scale'].b:
            self.set_weight(coreml_node_layer.name, "scale", np.array(coreml_node_bn.gamma.floatValue, dtype=np.float32))

        if IR_node.attr['bias'].b:
            self.set_weight(coreml_node_layer.name, "bias", np.array(coreml_node_bn.beta.floatValue, dtype=np.float32))



        gamma, beta = None, None
        if IR_node.attr['scale'].b:
            gamma = np.array(coreml_node_bn.gamma.floatValue, dtype=np.float32)
        if IR_node.attr['bias'].b:
            beta =  np.array(coreml_node_bn.beta.floatValue, dtype=np.float32)

        mean = np.array(coreml_node_bn.mean.floatValue)
        variance  =  np.array(coreml_node_bn.variance.floatValue)

        gamma = np.ones(mean.shape) if gamma is None else gamma
        beta = np.zeros(mean.shape) if beta is None else beta

        # compute adjusted parameters
        # Reference: parameter transformation https://github.com/apple/coremltools/issues/153
        f = 1.0 / np.sqrt(variance +  coreml_node_bn.epsilon)
        gamma1 = gamma*f
        beta1 = beta - gamma*mean*f
        mean[:] = 0.0 #mean
        variance[:] = 1.0 - .00001 #stddev

        # convert type because of tensorflow
        gamma1 = gamma1.astype(np.float32)
        beta1 = beta1.astype(np.float32)
        mean = mean.astype(np.float32)
        variance = variance.astype(np.float32)


        if IR_node.attr['scale'].b:
            self.set_weight(coreml_node_layer.name, "scale", gamma1)

        if IR_node.attr['bias'].b:
            self.set_weight(coreml_node_layer.name, "bias", beta1)

        # mean
        self.set_weight(coreml_node_layer.name, "mean", mean)

        # var
        self.set_weight(coreml_node_layer.name, "var", variance)

    def rename_scale(self, coreml_node):


        IR_node = self.IR_graph.node.add()

        coreml_node_layer = coreml_node.layer
        coreml_node_scale = coreml_node_layer.scale



        # name, op
        CoremlParser._copy_and_reop(coreml_node, IR_node, "Scale")

        # input edge
        self.convert_inedge(coreml_node, IR_node)

        # bias
        IR_node.attr['use_bias'].b = coreml_node_scale.hasBias

        IR_node.attr['scale'].b = True


        self.set_weight(coreml_node_layer.name, "scale", np.array(coreml_node_scale.scale.floatValue).astype(np.float32))
        self.set_weight(coreml_node_layer.name, "scale_mean", np.zeros_like(coreml_node_scale.scale.floatValue).astype(np.float32))
        self.set_weight(coreml_node_layer.name, "scale_var", np.ones_like(coreml_node_scale.scale.floatValue).astype(np.float32))

        self.set_weight(coreml_node_layer.name, "shapeScale", coreml_node_scale.shapeScale[0])



        if IR_node.attr['use_bias'].b:
            self.set_weight(coreml_node_layer.name, "bias", np.array(coreml_node_scale.bias.floatValue).astype(np.float32))
            self.set_weight(coreml_node_layer.name, "shapeBias", coreml_node_scale.shapeBias[0])

    def rename_Pooling(self, coreml_node):


        IR_node = self.IR_graph.node.add()

        coreml_node_layer = coreml_node.layer
        coreml_node_pool = coreml_node_layer.pooling





        # name, op
        CoremlParser._copy_and_reop(coreml_node, IR_node, "Pool")

        # input edge
        self.convert_inedge(coreml_node, IR_node)

        kwargs = {}

        # MAX = 0, AVERAGE = 1, L2 = 2
        if coreml_node_pool.type == 0:
            kwargs['pooling_type'] = 'MAX'
        elif coreml_node_pool.type == 1:
            kwargs['pooling_type'] = 'AVG'
        elif coreml_node_pool.type == 2:
            kwargs['pooling_type'] = 'L2'



        is_global = coreml_node_pool.globalPooling


        if is_global:
            kwargs['global_pooling'] = True
            kwargs['global_pooling_coreml'] = True
            kwargs['shape_coreml'] = [self.shape_dict[coreml_node_layer.name][3], self.shape_dict[coreml_node_layer.name][4], self.shape_dict[coreml_node_layer.name][2] ]


        # padding
        self._convert_padding(coreml_node, IR_node)

        # strides
        # [1, sd, sh, sw, 1]
        kwargs['strides'] = [1] + list(coreml_node_pool.stride) + [1]


        # window_shape
        # [1, pd, ph, pw, 1]
        kwargs['kernel_shape'] = [1] + list(coreml_node_pool.kernelSize) + [1]




        assign_IRnode_values(IR_node, kwargs)




    def rename_Softmax(self, coreml_node):
        IR_node = self.IR_graph.node.add()

        # name, op
        CoremlParser._copy_and_reop(coreml_node, IR_node, 'Softmax')

        # input edge
        self.convert_inedge(coreml_node, IR_node)

    def rename_Concatenate(self, source_node):
        IR_node = self._convert_merge(source_node, 'Concat')


    def rename_Flatten(self, source_node):
        IR_node = self.IR_graph.node.add()

        # name, op
        CoremlParser._copy_and_reop(source_node, IR_node, 'Flatten')

        # input edge
        self.convert_inedge(source_node, IR_node)


    def rename_innerProduct(self, source_node):
        IR_node = self.IR_graph.node.add()

        # name, op
        CoremlParser._copy_and_reop(source_node, IR_node, "FullyConnected")

        # input edge
        self.convert_inedge(source_node, IR_node)

        source_node_layer = source_node.layer
        source_node_inner = source_node_layer.innerProduct


        # units
        IR_node.attr['units'].i = source_node_inner.outputChannels

        # use_bias
        IR_node.attr['use_bias'].b = source_node_inner.hasBias

        # weights
        self.set_weight(source_node_layer.name, 'weights', np.array(source_node_inner.weights.floatValue).astype(np.float32).reshape( source_node_inner.outputChannels,  source_node_inner.inputChannels).transpose() )
        if IR_node.attr['use_bias'].b:
            self.set_weight(source_node_layer.name, 'bias', np.array(source_node_inner.bias.floatValue).astype(np.float32) )
        # change to single because of the tf matmul
        
        # in features
        IR_node.attr['in_features'].i = source_node_inner.inputChannels

    def rename_Padding(self, source_node):
        IR_node = self.IR_graph.node.add()

        # name, op
        self._convert_padding_api(source_node, IR_node)




from collections import OrderedDict

def parse_prototxt(protofile):
    def line_type(line):
        if line.find(':') >= 0:
            return 0
        elif line.find('{') >= 0:
            return 1
        return -1

    def parse_block(fp):
        block = OrderedDict()
        line = fp.readline().strip()
        while line != '}':
            ltype = line_type(line)
            if ltype == 0: # key: value
                #print line
                line = line.split('#')[0]
                key, value = line.split(':')
                key = key.strip()
                value = value.strip().strip('"')
                if key in block.keys():
                    if type(block[key]) == list:
                        block[key].append(value)
                    else:
                        block[key] = [block[key], value]
                else:
                    block[key] = value
            elif ltype == 1: # blockname {
                key = line.split('{')[0].strip()
                sub_block = parse_block(fp)
                block[key] = sub_block
            line = fp.readline().strip()
            line = line.split('#')[0]
        return block

    fp = open(protofile, 'r')
    props = OrderedDict()
    layers = []
    line = fp.readline()
    while line != '':
        line = line.strip().split('#')[0]
        if line == '':
            line = fp.readline()
            continue
        ltype = line_type(line)
        if ltype == 0: # key: value
            key, value = line.split(':')
            key = key.strip()
            value = value.strip().strip('"')
            if key in props.keys():
               if type(props[key]) == list:
                   props[key].append(value)
               else:
                   props[key] = [props[key], value]
            else:
                props[key] = value
        elif ltype == 1: # blockname {
            key = line.split('{')[0].strip()
            if key == 'layer':
                layer = parse_block(fp)
                layers.append(layer)
            else:
                props[key] = parse_block(fp)
        line = fp.readline()

    if len(layers) > 0:
        net_info = OrderedDict()
        net_info['props'] = props
        net_info['layers'] = layers
        return net_info
    else:
        return props

def is_number(s):
    try:
        float(s)
        return True
    except ValueError:
        return False



from __future__ import absolute_import
from setuptools import setup, find_packages
from io import open

# Get the long description from the README file
with open('README.md', encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='mmdnn',

    # Versions should comply with PEP440.  For a discussion on single-sourcing
    # the version across setup.py and the project code, see
    # https://packaging.python.org/en/latest/single_source_version.html
    version='0.2.4',

    description='Deep learning model converter, visualization and editor.',
    long_description=long_description,
    long_description_content_type='text/markdown',

    # The project's main homepage.
    url='https://github.com/Microsoft/MMdnn',

    # Author details
    author = 'System Research Group, Microsoft Research Asia',
    author_email='mmdnn_feedback@microsoft.com',

    # Choose your license
    license='MIT',

    # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        'Development Status :: 3 - Alpha',

        # Indicate who your project is intended for
        'Intended Audience :: Developers',
        'Intended Audience :: Education',
        'Intended Audience :: Science/Research',
        'Topic :: Scientific/Engineering :: Mathematics',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: Software Development :: Libraries',

        # Pick your license as you wish (should match "license" above)
        'License :: OSI Approved :: MIT License',

        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2, Python 3 or both.
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 3'
    ],

    # What does your project relate to?
    keywords='deep learning model converter visualization',

    # You can just specify the packages manually here if your project is
    # simple. Or you can use find_packages().
    packages=find_packages(),

    package_data={
        'mmdnn':['visualization/public/*',
                'visualization/*.json',
                'visualization/*.js',
                'visualization/*.html',
                'visualization/*.css']
    },

    # Alternatively, if you want to distribute just a my_module.py, uncomment
    # this:
    #   py_modules=["my_module"],

    # List run-time dependencies here.  These will be installed by pip when
    # your project is installed. For an analysis of "install_requires" vs pip's
    # requirements files see:
    # https://packaging.python.org/en/latest/requirements.html
    install_requires=[
        'numpy >= 1.15.0',
        'protobuf >= 3.6.0',
        'six >= 1.10.0',
        'pillow >= 3.1.0',
    ],

    # To provide executable scripts, use entry points in preference to the
    # "scripts" keyword. Entry points provide cross-platform support and allow
    # pip to create the appropriate form of executable for the target platform.
    entry_points={
        'console_scripts': [
            'mmconvert  = mmdnn.conversion._script.convert:_main',
            'mmdownload = mmdnn.conversion._script.extractModel:_main',
            'mmvismeta  = mmdnn.conversion.examples.tensorflow.vis_meta:_main',
            'mmtoir     = mmdnn.conversion._script.convertToIR:_main',
            'mmtocode   = mmdnn.conversion._script.IRToCode:_main',
            'mmtomodel  = mmdnn.conversion._script.dump_code:_main',
        ],
    },
)
